/*
 *  linux/drivers/char/pxa_gpio.c
 *
 *  Copyright (C) 2007 Florian Pantaleao, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This driver gives user access to GPIO control.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/pxa_gpio.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/signal.h>
#include <asm/arch/pxa-regs.h>
#include <asm/mach/irq.h>

#ifdef CONFIG_PXA27x
#define NR_PXA_GPIO 121
#endif

#ifndef NR_PXA_GPIO
#define NR_PXA_GPIO 80
#endif

struct gpio_use {
  pxa_gpio_use_t use;
  pid_t          user_pid;
};

struct pxa_gpio_context {
  int		   use_count;
  struct semaphore sem;
  struct gpio_use  gpio[NR_PXA_GPIO];
};

static struct pxa_gpio_context pg_ctx;

#if 0
static void dump_inuse(char *str)
{
  int ii;
  printk("dump_inuse %s:\n", str);
  for (ii=0; ii<NR_PXA_GPIO; ii++)
    printk("% 3d %d % 3d\n",
	   ii, pg_ctx.gpio[ii].use, pg_ctx.gpio[ii].user_pid);
  printk("\n");
}
#endif

int pxa_gpio_use(unsigned int gpio, pid_t pid, pxa_gpio_use_t use)
{
  int ret = 0;

  if (gpio >= NR_PXA_GPIO) return -1;

  struct gpio_use *pgpio = &pg_ctx.gpio[gpio];
  down(&pg_ctx.sem);
  if (pgpio->use != PXA_GPIO_NONE)
    ret = -1;
  else {
    pgpio->use = use;
    pgpio->user_pid = pid;
  }
  up(&pg_ctx.sem);

  return ret;
}

int pxa_gpio_unuse(unsigned int gpio, pid_t pid, pxa_gpio_use_t use)
{
  int ret = 0;

  if (gpio >= NR_PXA_GPIO) return -1;

  struct gpio_use *pgpio = &pg_ctx.gpio[gpio];
  down(&pg_ctx.sem);
  if ((pgpio->use == use) && (pgpio->user_pid == pid)) {
    pgpio->use = PXA_GPIO_NONE;
    pxa_gpio_mode(gpio | GPIO_IN);
  }
  else
    ret = -1;
  up(&pg_ctx.sem);

  return ret;
}



/*
 * User space driver interface.
 */

static int pxa_gpio_open(struct inode *inode, struct file *filp)
{
  int ret = 0;

  pg_ctx.use_count++;
  filp->private_data = &pg_ctx;

  return ret;
}

static int pxa_gpio_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
  int ret = 0;
	
  switch(cmd)
    {
    case PXAIOCGPIOADD: {
      struct pxa_gpio_add gpio_add;

      if (copy_from_user(&gpio_add, (void*) arg, sizeof(gpio_add))) {
	ret = -EFAULT;
	break;
      }
 
      /* sanity check */
      if (gpio_add.gpio >= NR_PXA_GPIO) {
	ret = -EINVAL;
	break;
      }
      pxa_gpio_use_t use;
      if (gpio_add.input) use = PXA_GPIO_INPUT;
      else use = PXA_GPIO_OUTPUT;
      if (pxa_gpio_use(gpio_add.gpio, current->pid, use)) {
	ret = -EBUSY;
	break;
      }

      /* Process add request */
      if (gpio_add.input) {
	/* input mode */
	pxa_gpio_mode(gpio_add.gpio | GPIO_IN);
	gpio_add.init_state = GPLR(gpio_add.gpio) & GPIO_bit(gpio_add.gpio);
	if (copy_to_user((void*) arg, &gpio_add, sizeof(gpio_add)))
	  ret = -EFAULT;
      }
      else {
	/* output mode */
	pxa_gpio_mode(gpio_add.gpio | GPIO_OUT);
	if (gpio_add.init_state)
	  GPSR(gpio_add.gpio) |= GPIO_bit(gpio_add.gpio);
	else
	  GPCR(gpio_add.gpio) |= GPIO_bit(gpio_add.gpio);
      }

      break;
    }

    case PXAIOCGPIOREM: {
      struct pxa_gpio_rem gpio_rem;

      if (copy_from_user(&gpio_rem, (void*) arg, sizeof(gpio_rem))) {
	ret = -EFAULT;
	break;
      }
 
      /* sanity check */
      if (gpio_rem.gpio >= NR_PXA_GPIO) {
	ret = -EINVAL;
	break;
      }

      /* Process remove request */
      pxa_gpio_use_t use;
      if (gpio_rem.input) use = PXA_GPIO_INPUT;
      else use = PXA_GPIO_OUTPUT;
      if (pxa_gpio_unuse(gpio_rem.gpio, current->pid, use)) {
	ret = -EBUSY;
      }
      break;
    }

    case PXAIOCGPIOSET: {
      unsigned int gpio;

      if (copy_from_user(&gpio, (void*) arg, sizeof(gpio))) {
	ret = -EFAULT;
	break;
      }
 
      /* sanity check */
      if (gpio >= NR_PXA_GPIO) {
	ret = -EINVAL;
	break;
      }
      if ((pg_ctx.gpio[gpio].use != PXA_GPIO_OUTPUT) ||
	  (pg_ctx.gpio[gpio].user_pid != current->pid)) {
	ret = -EINVAL;
	break;
      }
      
      /* Process set request */
      GPSR(gpio) |= GPIO_bit(gpio);
      break;
    }

    case PXAIOCGPIOCLR: {
      unsigned int gpio;

      if (copy_from_user(&gpio, (void*) arg, sizeof(gpio))) {
	ret = -EFAULT;
	break;
      }
 
      /* sanity check */
      if (gpio >= NR_PXA_GPIO) {
	ret = -EINVAL;
	break;
      }
      if ((pg_ctx.gpio[gpio].use != PXA_GPIO_OUTPUT) ||
	  (pg_ctx.gpio[gpio].user_pid != current->pid)) {
	ret = -EINVAL;
	break;
      }
      
      /* Process clear request */
      GPCR(gpio) |= GPIO_bit(gpio);
      break;
    }

    case PXAIOCGPIOGET: {
      struct pxa_gpio_get gpio_get;

      if (copy_from_user(&gpio_get, (void*) arg, sizeof(gpio_get))) {
	ret = -EFAULT;
	break;
      }

      /* sanity check */
      if (gpio_get.gpio > NR_PXA_GPIO) {
	ret = -EINVAL;
	break;
      }
      if ((pg_ctx.gpio[gpio_get.gpio].use != PXA_GPIO_INPUT) ||
	  (pg_ctx.gpio[gpio_get.gpio].user_pid != current->pid)) {
	ret = -EINVAL;
	break;
      }

      /* Process get request */
      gpio_get.state = GPLR(gpio_get.gpio) & GPIO_bit(gpio_get.gpio);
      if (copy_to_user((void*) arg, &gpio_get, sizeof(gpio_get)))
	ret = -EFAULT;
      break;
    }

    default:
      ret = -EINVAL;
    }

  return ret;
}

static int pxa_gpio_release(struct inode *inode, struct file *filp)
{
  /* free all used GPIOs if PXAIOCGPIOREM has not been called */
  int ii;
  struct gpio_use  *pgpio;
  down(&pg_ctx.sem);
  for (ii=0, pgpio=pg_ctx.gpio; ii<NR_PXA_GPIO; ii++, pgpio++) {
    if (((pgpio->use == PXA_GPIO_INPUT) || (pgpio->use == PXA_GPIO_OUTPUT)) &&
	(pgpio->user_pid == current->pid)) {
      pgpio->use = PXA_GPIO_NONE;
      pxa_gpio_mode(ii | GPIO_IN);
    }
  }
  up(&pg_ctx.sem);

  pg_ctx.use_count--;
  
  return 0;
}

static struct file_operations pxa_gpio_fops = {
	owner:		THIS_MODULE,
	open:		pxa_gpio_open,
	release:	pxa_gpio_release,
	ioctl:		pxa_gpio_ioctl,
};


/*
 * This is an unofficial PXA GPIQ driver considered as a miscdevice:
 *   10 char        Non-serial mice, misc features
 *                  18 = /dev/pxa-gpio
 */
static struct miscdevice pxa_gpio_dev = {
	minor:	PXA_GPIO_MINOR,
	name:	"pxa-gpio",
	fops:	&pxa_gpio_fops,
};


/*
 * Initialization.
 */
static int __init pxa_gpio_init(void)
{
  init_MUTEX(&pg_ctx.sem);

#ifndef PXA_GPIO_UNUSED
#error "Unused GPIOs list must be defined"
#endif

  /* GPIO usage: default=all GPIOs used by system then board specific gpio release */
  struct gpio_use  *pgpio;
  for (pgpio=pg_ctx.gpio; pgpio<&pg_ctx.gpio[NR_PXA_GPIO]; pgpio++) {
    pgpio->use = PXA_GPIO_SYSTEM;
    pgpio->user_pid = 0;
  }
  int gpio_list[] = PXA_GPIO_UNUSED;
  int nr_unused_gpio = sizeof(gpio_list)/sizeof(gpio_list[0]);
  int ii;
  for (ii=0; ii<nr_unused_gpio; ii++) 
    pxa_gpio_unuse(gpio_list[ii], 0, PXA_GPIO_SYSTEM);
  return misc_register(&pxa_gpio_dev);
}

static void __exit pxa_gpio_exit(void)
{
  misc_deregister(&pxa_gpio_dev);
}

module_init(pxa_gpio_init);
module_exit(pxa_gpio_exit);

MODULE_AUTHOR("Florian Pantaleao <fpantaleao@mobisensesystems.com>");
MODULE_DESCRIPTION("PXA GPIO driver");
MODULE_LICENSE("GPL");
