/*
 *  linux/drivers/char/pxa_pwm.c
 *
 *  Copyright (C) 2007 Florian Pantaleao, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This driver uses the Pulse Width Modulator (PWM) controller integrated in
 * PXA processors to generate accurate signals with configurable duty cycle
 * and frequency.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/pxa_gpio.h>
#include <linux/pxa_pwm.h>
#include <linux/miscdevice.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/signal.h>
#include <asm/arch/pxa-regs.h>
#include <asm/mach/irq.h>

#if NR_PXA_PWM < 4
#error "pb nr_pxa_pwm"
#endif

#define PXA_PWM_RES                 10
#define PXA_PWM_INPUT_CLOCK   13000000
#define PXA_PWM_MAX_PRESCALE      0x3f

struct pxa_pwm_data {
  unsigned int  idle;
  unsigned int  gpio;
  unsigned int  alt_fn;
  unsigned int  duty_shift;
  pid_t         pid;
  u32           *pwmcr, *pwmdcr, *pwmpcr;
};

struct pxa_pwm_context {
  int			use_count;
  struct semaphore      sem;
  struct pxa_pwm_data   pwm[NR_PXA_PWM];
};

static struct pxa_pwm_context pp_ctx;


static void pxa_pwm_set_freq(unsigned int f_hz, struct pxa_pwm_data *pwm)
{
  unsigned int res = PXA_PWM_RES;
  int prescale;
  while (res >= 0) {
    prescale = (int) (PXA_PWM_INPUT_CLOCK / (f_hz * (1U << res))) - 1;
    if (prescale >= 0) break;
    res--;
  }
  pwm->duty_shift = PXA_PWM_RES - res;
  if (prescale > PXA_PWM_MAX_PRESCALE) prescale = PXA_PWM_MAX_PRESCALE;
  *(pwm->pwmcr)  = prescale;
  *(pwm->pwmpcr) = (1U << (PXA_PWM_RES - pwm->duty_shift)) - 1;
}

static inline void pxa_pwm_set_duty(unsigned int duty, struct pxa_pwm_data *pwm)
{
  *(pwm->pwmdcr) = (duty >> pwm->duty_shift);
}

static void pxa_pwm_unuse(struct pxa_pwm_data *pwm)
{
  pwm->idle   = 1;
  *(pwm->pwmcr)  = 0; /* no prescale, graceful shutdown enabled*/
  *(pwm->pwmdcr) = 0; /* output deasserted */
  *(pwm->pwmpcr) = (1U << PXA_PWM_RES) - 1;
  pxa_gpio_unuse(pwm->gpio, current->pid, PXA_GPIO_PWM);
}

/*
 * User space driver interface.
 */

static int pxa_pwm_open(struct inode *inode, struct file *filp)
{
  int ret = 0;

  if (pp_ctx.use_count == 0) {
    /* register preset */
    struct pxa_pwm_data *pwm;
    for (pwm=pp_ctx.pwm; pwm<&pp_ctx.pwm[NR_PXA_PWM]; pwm++)
      pxa_pwm_unuse(pwm);
    
    /* turn on PWM controller:
       - In PXA255, PWM0 and PWM1 can be independently enabled with
         CKEN0 and CKEN1 respectively.
       - In PXA270, all PWMs are enabled with CKEN0 or CKEN1 set.
       => set both for compatibility with both processors */
    pxa_set_cken(CKEN0_PWM0 | CKEN1_PWM1, 1);
  }

  pp_ctx.use_count++;
  filp->private_data = &pp_ctx;

  return ret;
}


static int pxa_pwm_ioctl(struct inode *inode, struct file *filp,
	unsigned int cmd, unsigned long arg)
{
  int ret = 0;
  struct pxa_pwm_data *pwm;

  switch(cmd)
    {
    case PXAIOCPWMADD: {
      struct pxa_pwm_add pwm_add;

      if (copy_from_user(&pwm_add, (void*) arg, sizeof(pwm_add))) {
	ret = -EFAULT;
	break;
      }
 
      /* sanity check */
      if ((pwm_add.nr >= NR_PXA_PWM) || (pwm_add.f_hz == 0)){
	ret = -EINVAL;
	break;
      }

      pwm = &(pp_ctx.pwm[pwm_add.nr]);
      down(&pp_ctx.sem);
      if (!(pwm->idle)) {
	ret = -EBUSY;
	goto pxaiocpwmadd_release_mutex;
      }
      if (pxa_gpio_use(pwm->gpio, current->pid, PXA_GPIO_PWM)) {
	ret = -EBUSY;
	goto pxaiocpwmadd_release_mutex;
      }

      /* Process add request */
      pwm->idle = 0;
      pwm->pid  = current->pid;
      pxa_pwm_set_freq(pwm_add.f_hz, pwm);
      pxa_pwm_set_duty(pwm_add.duty, pwm);
      pxa_gpio_mode(pwm->gpio | pwm->alt_fn);

    pxaiocpwmadd_release_mutex:
      up(&pp_ctx.sem);
      break;
    }

    case PXAIOCPWMREM: {
      unsigned int pwm_nr;

      if (copy_from_user(&pwm_nr, (void*) arg, sizeof(pwm_nr))) {
	ret = -EFAULT;
	break;
      }
 
      /* sanity check */
      if (pwm_nr >= NR_PXA_PWM) {
	ret = -EINVAL;
	break;
      }
      pwm = &pp_ctx.pwm[pwm_nr];
      down(&pp_ctx.sem);
      if ((pwm->idle) || (pwm->pid != current->pid)) {
	ret = -EINVAL;
	goto pxaiocpwmrem_release_mutex;
      }

      /* Process remove request */
      pxa_pwm_unuse(pwm);

      pxaiocpwmrem_release_mutex:
      up(&pp_ctx.sem);
      break;
    }

    case PXAIOCPWMDUTY: {
      int ii;
      struct pxa_pwm_chg pwm_chg;

      if (copy_from_user(&pwm_chg, (void*) arg, sizeof(pwm_chg))) {
	ret = -EFAULT;
	break;
      }

      /* sanity check */
      if (pwm_chg.nr_chg > NR_PXA_PWM) {
	ret = -EINVAL;
	break;
      }
      down(&pp_ctx.sem);
      for (ii=0; ii<pwm_chg.nr_chg; ii++) {
	if (pwm_chg.pwm[ii].nr > NR_PXA_PWM) {
	  ret = -EINVAL;
	  break;
	}
	pwm = &pp_ctx.pwm[pwm_chg.pwm[ii].nr];
	if ((pwm->idle) || (pwm->pid != current->pid)) {
	  ret = -EINVAL;
	  break;
	}
	pxa_pwm_set_duty(pwm_chg.pwm[ii].val, pwm);
      }
      up(&pp_ctx.sem);
      break;
    }

    default:
      ret = -EINVAL;
    }

  return ret;
}


static int pxa_pwm_release(struct inode *inode, struct file *filp)
{
  /* free all used GPIOs if PXAIOCPWMREM has not been called */
  struct pxa_pwm_data *pwm;
  down(&pp_ctx.sem);
  for (pwm=pp_ctx.pwm; pwm<&pp_ctx.pwm[NR_PXA_PWM]; pwm++) {
    if (!(pwm->idle) && (pwm->pid == current->pid)) {
      pxa_pwm_unuse(pwm);
    }
  }
  up(&pp_ctx.sem);

  pp_ctx.use_count--;
  if (pp_ctx.use_count == 0) {
    /* Turn off PWM controller at last call */
    pxa_set_cken(CKEN0_PWM0 | CKEN1_PWM1, 0);
  }
  
  return 0;
}

static struct file_operations pxa_pwm_fops = {
	owner:		THIS_MODULE,
	open:		pxa_pwm_open,
	release:	pxa_pwm_release,
	ioctl:		pxa_pwm_ioctl,
};


/*
 * This is an unofficial PXA PWM driver considered as a miscdevice:
 *   10 char        Non-serial mice, misc features
 *                  19 = /dev/pxa-pwm
 */
static struct miscdevice pxa_pwm_dev = {
	minor:	PXA_PWM_MINOR,
	name:	"pxa-pwm",
	fops:	&pxa_pwm_fops,
};


/*
 * Initialization.
 */
static int __init pxa_pwm_init(void)
{
  init_MUTEX(&pp_ctx.sem);
  pxa_set_cken(CKEN0_PWM0 | CKEN1_PWM1, 0);
  pp_ctx.pwm[0].pwmcr  = (u32*) &(PWM_CTRL0);
  pp_ctx.pwm[0].pwmdcr = (u32*) &(PWM_PWDUTY0);
  pp_ctx.pwm[0].pwmpcr = (u32*) &(PWM_PERVAL0);
  pp_ctx.pwm[1].pwmcr  = (u32*) &(PWM_CTRL1);
  pp_ctx.pwm[1].pwmdcr = (u32*) &(PWM_PWDUTY1);
  pp_ctx.pwm[1].pwmpcr = (u32*) &(PWM_PERVAL1);
#if defined(CONFIG_PXA27x)
  pp_ctx.pwm[2].pwmcr  = (u32*) &(PWM_CTRL2);
  pp_ctx.pwm[2].pwmdcr = (u32*) &(PWM_PWDUTY2);
  pp_ctx.pwm[2].pwmpcr = (u32*) &(PWM_PERVAL2);
  pp_ctx.pwm[3].pwmcr  = (u32*) &(PWM_CTRL3);
  pp_ctx.pwm[3].pwmdcr = (u32*) &(PWM_PWDUTY3);
  pp_ctx.pwm[3].pwmpcr = (u32*) &(PWM_PERVAL3);
#endif

#if !defined(PXA_PWM_GPIO) || !defined(PXA_PWM_GPIO_ALT_FN)
#error "PWM GPIOs list and alternate functions must be defined"
#endif
  unsigned int gpio_list[] = PXA_PWM_GPIO;
  unsigned int alt_fn_list[] = PXA_PWM_GPIO_ALT_FN;
  int ii;
  for (ii=0; ii<NR_PXA_PWM; ii++) {
    pp_ctx.pwm[ii].gpio   = gpio_list[ii];
    pp_ctx.pwm[ii].alt_fn = alt_fn_list[ii];
  }

  return misc_register(&pxa_pwm_dev);
}

static void __exit pxa_pwm_exit(void)
{
  misc_deregister(&pxa_pwm_dev);
}

module_init(pxa_pwm_init);
module_exit(pxa_pwm_exit);

MODULE_AUTHOR("Florian Pantaleao <fpantaleao@mobisensesystems.com>");
MODULE_DESCRIPTION("PXA PWM driver");
MODULE_LICENSE("GPL");
