/*
 *  linux/drivers/char/pxa_servo.c
 *
 *  Copyright (C) 2007 Florian Pantaleao, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This driver uses a PXA OS timer and FIQ to generate submicrosecond accurate
 * pulses for RC servo control. It is inspired form previous similar projects:
 * Nivolas Pitre and Jangwon Boo, real-time scheduler at pabr.org.
 *
 * Kernel settings:
 * CONFIG_FIQ must be defined
 * include/asm/arch/irqs.h must have the line #define FIQ_START 0
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/pxa_gpio.h>
#include <linux/pxa_servo.h>
#include <linux/miscdevice.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/fiq.h>
#include <asm/signal.h>
#include <asm/arch/pxa-regs.h>
#include <asm/mach/irq.h>

#define PXA_SERVO_WAIT           0
#define PXA_SERVO_PULSE          1
#define PXA_SERVO_TIMER_COMP  (-6) /* clock ticks to compensate timer error */
#define PXA_SERVO_STACK_SIZE   512
#define PXA_SERVO_PERIOD_US  20000

struct pxa_servo_data {
  unsigned int  gpio;
  pid_t          pid;
  u32           idle;
  u32           gpio_bit;
  volatile u32 *gpcr;
  volatile u32 *gpsr;
  int           pulse;
};

struct pxa_servo_context {
  int			use_count;
  struct semaphore      sem;
  int                   state;		/* PXA_SERVO_WAIT or PXA_SERVO_PULSE */
  int                   no_servo;
  u32                   pulse_start, seq_start;
  struct pxa_servo_data servo[NR_PXA_SERVO];
  struct pxa_servo_data *pservo, *end_servo;
  unsigned char         fiq_stack[PXA_SERVO_STACK_SIZE];
};

static struct pxa_servo_context ps_ctx;

/* !!! OSCR clock frequency is 3.25MHz on PXA27x, 3.6864MHz on PXA25x */
#if defined(CONFIG_PXA27x)
/* 1s = 3.25 OSCR clock ticks = 13/4 */
#define USEC2TICK(x)  (((x) * 13) >> 2)
#else
/* 1s = 3.6864 OSCR clock ticks = 59/16 */
#define USEC2TICK(x)  (((x) * 59) >> 4)
#endif

static struct fiq_handler fh = {
 name: "pxa-servo"
};


#if 0
static void dump_servo_data(char *str)
{
  struct pxa_servo_data *pservo;
  printk("dump_servo_data %s:\n", str);
  for (pservo=ps_ctx.servo; pservo<&ps_ctx.servo[NR_PXA_SERVO]; pservo++)
    printk("% 3d % 3d %d % 4d\n",
	   pservo->gpio, pservo->pid, pservo->idle, pservo->pulse);
  printk("\n");
}
#endif

static void pxa_servo_disable_fiq(unsigned int irq)
{
  disable_fiq(irq);
#if PXA_INTERNAL_IRQS > 32
  if ( irq >= PXA_IRQ(32) )
    ICLR2 &= ~(1 << (irq - 32 + PXA_IRQ_SKIP));
  else
#endif
    ICLR &= ~(1 << (irq + PXA_IRQ_SKIP));
}

static void pxa_servo_enable_fiq(unsigned int irq)
{
#if PXA_INTERNAL_IRQS > 32
  if ( irq >= PXA_IRQ(32) )
    ICLR2 |= (1 << (irq - 32 + PXA_IRQ_SKIP));
  else
#endif
    ICLR |= (1 << (irq + PXA_IRQ_SKIP));
  enable_fiq(irq);
}

void pxa_servo_fiq_process(void)
{
  if (ps_ctx.state == PXA_SERVO_PULSE) {
    /* clear output */
    *(ps_ctx.pservo->gpcr) |= ps_ctx.pservo->gpio_bit;
    ps_ctx.pservo++;
  }

 search_servo:
  while (ps_ctx.pservo < ps_ctx.end_servo) {
    if (ps_ctx.pservo->idle) ps_ctx.pservo++;
    else break;
  }

  if (ps_ctx.pservo < ps_ctx.end_servo) {
    u32 pulse_start;
    /* set output */
    *(ps_ctx.pservo->gpsr) |= ps_ctx.pservo->gpio_bit;
    pulse_start = OSCR;
    /* Wait until pulse end */
    OSMR2 = pulse_start + (u32)ps_ctx.pservo->pulse + PXA_SERVO_TIMER_COMP;
    ps_ctx.state = PXA_SERVO_PULSE;
    if (ps_ctx.no_servo) {
      ps_ctx.no_servo = 0;
      ps_ctx.seq_start = pulse_start;
    }
  }
  else {
    u32 delay = 0;
    if (ps_ctx.no_servo) {
      delay = USEC2TICK(PXA_SERVO_PERIOD_US);
    }
    else {
      u32 seq_duration = OSCR - ps_ctx.seq_start;
      if (seq_duration < USEC2TICK(PXA_SERVO_PERIOD_US-1000)) {
	/* worth waiting */
	delay = USEC2TICK(PXA_SERVO_PERIOD_US) - seq_duration;
      }
    }
    /* Get ready for next sequence */
    ps_ctx.pservo = ps_ctx.servo;
    ps_ctx.no_servo = 1;
    if (delay) {
      /* Wait until next match */
      OSMR2 = OSCR + (u32)delay + PXA_SERVO_TIMER_COMP;
      ps_ctx.state = PXA_SERVO_WAIT;
    }
    else {
      /* Restart sequence immediatly */
      goto search_servo;
    }
  }
}

void pxa_servo_fiq_handler_end(void);

void __attribute__((naked)) pxa_servo_fiq_handler(void)
{
  __asm__ volatile (
		"mov     r12, #4                 @ load OSSR_M2 value        \n\
                 str     r12, [r8, #0x14]        @ Clear interrupt in OSSR   \n\
                 stmdb   sp!, {r0-r3,lr}         @ Save registers            \n\
                 mov lr, pc                      @ Call pxa_servo_fiq_process \n\
                 mov pc, r9                                                  \n\
                 ldmia   sp!, {r0-r3,lr}         @ Restore registers         \n\
                 subs    pc, lr, #4                                          \n\
                 pxa_servo_fiq_handler_end:                                   \n\
                ");
}


/*
 * User space driver interface.
 */

static int pxa_servo_open(struct inode *inode, struct file *filp)
{
  int ii;
  struct pt_regs FIQ_regs;
  int ret = 0;

  if (ps_ctx.use_count == 0) {
    /* Request timer FIQ at first call */
    ps_ctx.state = PXA_SERVO_WAIT;
    ps_ctx.pservo = ps_ctx.servo;
    ps_ctx.no_servo = 1;
    for (ii=0; ii<NR_PXA_SERVO; ii++) ps_ctx.servo[ii].idle = 1;

    if(claim_fiq(&fh)) {
      ret = -EBUSY;
      goto open_end;
    }

    /* set FIQ registers */
    FIQ_regs.ARM_r8 = (u32)&OSMR0;
    FIQ_regs.ARM_r9 = (u32)pxa_servo_fiq_process;
    FIQ_regs.ARM_sp = (u32)&ps_ctx.fiq_stack[PXA_SERVO_STACK_SIZE];
    set_fiq_handler(pxa_servo_fiq_handler,
		    pxa_servo_fiq_handler_end-pxa_servo_fiq_handler);
    set_fiq_regs(&FIQ_regs);
    pxa_servo_enable_fiq(IRQ_OST2);

    /* enable OSMR2 FIQ */
    OSSR = OSSR_M2;
    OSMR2 = OSCR + USEC2TICK(1000);
    OIER |= OIER_E2;
  }

  ps_ctx.use_count++;
  filp->private_data = &ps_ctx;

open_end:
  return ret;
}


static int pxa_servo_ioctl(struct inode *inode, struct file *filp,
	unsigned int cmd, unsigned long arg)
{
  int ret = 0;
  struct pxa_servo_data *pservo;

  switch(cmd)
    {
    case PXAIOCSRVADD: {
      struct pxa_servo_add srv_add;
      int ii;

      if (copy_from_user(&srv_add, (void*) arg, sizeof(struct pxa_servo_add))) {
	ret = -EFAULT;
	break;
      }
 
      /* sanity check */
      if ((srv_add.delay < PXA_SERVO_MIN_DELAY_US) ||
	  (srv_add.delay > PXA_SERVO_MAX_DELAY_US)) {
	ret = -EINVAL;
	break;
      }

      down(&ps_ctx.sem);
      /* find a free channel */
      for (ii=0; ii<NR_PXA_SERVO; ii++) {
	if (ps_ctx.servo[ii].idle) break;
      }
      if (ii == NR_PXA_SERVO) {
	ret = -EBUSY;
	goto pxaiocsrvadd_release_mutex;
      }
      /* reserve GPIO for servo output */
      if (pxa_gpio_use(srv_add.gpio, current->pid, PXA_GPIO_SERVO)) {
	ret = -EBUSY;
	goto pxaiocsrvadd_release_mutex;
      }

      /* Everything OK => Process add request */
      srv_add.id = ii;
      pxa_gpio_mode(srv_add.gpio | GPIO_OUT);
      pservo = &(ps_ctx.servo[ii]);
      pservo->gpio     = srv_add.gpio;
      pservo->pid      = current->pid;
      pservo->gpio_bit = GPIO_bit(srv_add.gpio);
      pservo->gpcr     = (u32*) &(GPCR(srv_add.gpio));
      pservo->gpsr     = (u32*) &(GPSR(srv_add.gpio));
      pservo->pulse    = USEC2TICK(srv_add.delay);
      pservo->idle     = 0;

      /* return servo id */
      if (copy_to_user((void*) arg, &srv_add, sizeof(struct pxa_servo_add)))
	ret = -EFAULT;

    pxaiocsrvadd_release_mutex:
      up(&ps_ctx.sem);

      break;
    }

    case PXAIOCSRVREM: {
      struct pxa_servo_rem srv_rem;

      if (copy_from_user(&srv_rem, (void*) arg, sizeof(struct pxa_servo_rem))) {
	ret = -EFAULT;
	break;
      }
 
      /* sanity check */
      if (srv_rem.id >= NR_PXA_SERVO) {
	ret = -EINVAL;
	break;
      }
      pservo = &ps_ctx.servo[srv_rem.id];
      if ((pservo->idle) || (pservo->pid != current->pid)) {
	ret = -EINVAL;
	break;
      }

      /* Process remove request */
      down(&ps_ctx.sem);
      pservo->idle = 1;
      if (pxa_gpio_unuse(pservo->gpio, current->pid, PXA_GPIO_SERVO))
	ret = -EINVAL;
      up(&ps_ctx.sem);

      break;
    }

    case PXAIOCSRVCHG: {
      struct pxa_servo_chg srv_chg;

      if (copy_from_user(&srv_chg, (void*) arg, sizeof(struct pxa_servo_chg))) {
	ret = -EFAULT;
	break;
      }

      /* sanity check */
      if (srv_chg.nb > NR_PXA_SERVO) {
	ret = -EINVAL;
	break;
      }

      struct pxa_servo_val srv_val[NR_PXA_SERVO];
      if (copy_from_user(&srv_val, (void*) srv_chg.val,
			 sizeof(struct pxa_servo_val) * srv_chg.nb)) {
	ret = -EFAULT;
	break;
      }

      int ii;
      for (ii=0; ii<srv_chg.nb; ii++) {
	/* sanity check */
	if ((srv_val[ii].id >= NR_PXA_SERVO) ||
	    (srv_val[ii].delay < PXA_SERVO_MIN_DELAY_US) ||
	    (srv_val[ii].delay > PXA_SERVO_MAX_DELAY_US)) {
	  ret = -EINVAL;
	  break;
	}
	pservo = &(ps_ctx.servo[srv_val[ii].id]);
	if (pservo->pid != current->pid) {
	  ret = -EINVAL;
	  break;
	}

	/* Change delay */
	pservo->pulse = USEC2TICK(srv_val[ii].delay);
      }

      break;
    }

    default:
      ret = -EINVAL;
    }

  return ret;
}


static int pxa_servo_release(struct inode *inode, struct file *filp)
{
  /* free all used GPIOs if PXAIOCSRVREM has not been called */
  struct pxa_servo_data *pservo;
  down(&ps_ctx.sem);
  for (pservo=ps_ctx.servo; pservo<&ps_ctx.servo[NR_PXA_SERVO]; pservo++) {
    if ((pservo->idle == 0) && (pservo->pid == current->pid)) {
      pservo->idle = 1;
      pxa_gpio_unuse(pservo->gpio, current->pid, PXA_GPIO_SERVO);
    }
  }
  up(&ps_ctx.sem);

  ps_ctx.use_count--;
  if (ps_ctx.use_count == 0) {
    /* Release timer FIQ at last call */
    OIER &= ~OIER_E2;
    pxa_servo_disable_fiq(IRQ_OST2);
    release_fiq(&fh);
  }
  
  return 0;
}

static struct file_operations pxa_servo_fops = {
	owner:		THIS_MODULE,
	open:		pxa_servo_open,
	release:	pxa_servo_release,
	ioctl:		pxa_servo_ioctl,
};


/*
 * This is an unofficial PXA FIQ driver considered as a miscdevice:
 *   10 char        Non-serial mice, misc features
 *                   17 = /dev/pxa-servo
 */
static struct miscdevice pxa_servo_dev = {
	minor:	PXA_SERVO_MINOR,
	name:	"pxa-servo",
	fops:	&pxa_servo_fops,
};


/*
 * Initialization.
 */
static int __init pxa_servo_init(void)
{
  ps_ctx.end_servo = &ps_ctx.servo[NR_PXA_SERVO];
  init_MUTEX(&ps_ctx.sem);
  return misc_register(&pxa_servo_dev);
}

static void __exit pxa_servo_exit(void)
{
	misc_deregister(&pxa_servo_dev);
}

module_init(pxa_servo_init);
module_exit(pxa_servo_exit);

MODULE_AUTHOR("Florian Pantaleao <fpantaleao@mobisensesystems.com>");
MODULE_DESCRIPTION("PXA FIQ driver for RC servomotor control");
MODULE_LICENSE("GPL");
