/*
 *  linux/drivers/char/pxa_timer.c
 *
 *  Copyright (C) 2007 Florian Pantaleao, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This driver uses extra timers of PXA270 operating system timer block to
 * generate fixed frequency outputs on CHOUT0 and CHOUT1.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/pxa_gpio.h>
#include <linux/pxa_timer.h>
#include <linux/miscdevice.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/signal.h>
#include <asm/arch/pxa-regs.h>
#include <asm/mach/irq.h>

#define PXA_TIMER_INPUT_CLOCK   13000000

struct pxa_timer_chout {
  unsigned int  idle;
  unsigned int  gpio;
  unsigned int  alt_fn;
  pid_t         pid;
  u32           *omcr, *osmr, *oscr;
};

struct pxa_timer_context {
  int			 use_count;
  struct semaphore       sem;
  struct pxa_timer_chout chout[NR_PXA_CHOUT];
};

static struct pxa_timer_context pt_ctx;


static void pxa_timer_set_freq(unsigned int f_hz,
			       unsigned int ratio,
			       struct pxa_timer_chout *chout)
{
  u32 counter_freq_hz;
  u32 tmp = *(chout->omcr) & ~OMCR_CRES;
  unsigned int period = ratio / f_hz;
  if (period > 1000) {
    /* counter resolution = 1 millisecond */
    *(chout->omcr) = tmp | 0x2;
    counter_freq_hz = 1000;
  }
  else {
    /* counter resolution = 1 microsecond */
    *(chout->omcr) = tmp | 0x4;
    counter_freq_hz = 1000000;
  }
  *(chout->osmr) = (counter_freq_hz * ratio) / (2 * f_hz) - 1;
  *(chout->oscr) = 0;
}

static void pxa_timer_unuse_chout(struct pxa_timer_chout *chout)
{
  chout->idle    = 1;
  *(chout->omcr) &= ~(OMCR_P | OMCR_R | OMCR_C);
  pxa_gpio_unuse(chout->gpio, current->pid, PXA_GPIO_CLOCK);
}

/*
 * User space driver interface.
 */

static int pxa_timer_open(struct inode *inode, struct file *filp)
{
  int ret = 0;

  if (pt_ctx.use_count == 0) {
    /* register preset */
    struct pxa_timer_chout *chout;
    for (chout=pt_ctx.chout; chout<&pt_ctx.chout[NR_PXA_CHOUT]; chout++)
      pxa_timer_unuse_chout(chout);
  }

  pt_ctx.use_count++;
  filp->private_data = &pt_ctx;

  return ret;
}


static int pxa_timer_ioctl(struct inode *inode, struct file *filp,
	unsigned int cmd, unsigned long arg)
{
  int ret = 0;
  struct pxa_timer_chout *chout;

  switch(cmd)
    {
    case PXAIOCTMR_ADDCO: {
      struct pxa_timer_add_chout add_chout;

      if (copy_from_user(&add_chout, (void*) arg, sizeof(add_chout))) {
	ret = -EFAULT;
	break;
      }
 
      /* sanity check */
      if ((add_chout.nr >= NR_PXA_CHOUT) ||
	  (add_chout.f_hz == 0) ||
	  (add_chout.ratio < 1)) {
	ret = -EINVAL;
	break;
      }

      chout = &(pt_ctx.chout[add_chout.nr]);
      down(&pt_ctx.sem);
      if (!(chout->idle)) {
	ret = -EBUSY;
	goto pxaioctmr_addco_release_mutex;
      }
      if (pxa_gpio_use(chout->gpio, current->pid, PXA_GPIO_CLOCK)) {
	ret = -EBUSY;
	goto pxaioctmr_addco_release_mutex;
      }

      /* Process add request */
      chout->idle = 0;
      chout->pid  = current->pid;
      u32 tmp = *(chout->omcr) & ~(OMCR_S0 | OMCR_S1);
      *(chout->omcr) = tmp | OMCR_P | OMCR_R | OMCR_C;
      pxa_timer_set_freq(add_chout.f_hz, add_chout.ratio, chout);
      pxa_gpio_mode(chout->gpio | chout->alt_fn);

    pxaioctmr_addco_release_mutex:
      up(&pt_ctx.sem);
      break;
    }

    case PXAIOCTMR_REMCO: {
      unsigned int chout_nr;

      if (copy_from_user(&chout_nr, (void*) arg, sizeof(chout_nr))) {
	ret = -EFAULT;
	break;
      }
 
      /* sanity check */
      if (chout_nr >= NR_PXA_CHOUT) {
	ret = -EINVAL;
	break;
      }
      chout = &pt_ctx.chout[chout_nr];
      down(&pt_ctx.sem);
      if ((chout->idle) || (chout->pid != current->pid)) {
	ret = -EINVAL;
	goto pxaioctmr_remco_release_mutex;
      }

      /* Process remove request */
      pxa_timer_unuse_chout(chout);

      pxaioctmr_remco_release_mutex:
      up(&pt_ctx.sem);
      break;
    }

    case PXAIOCTMR_FREQCO: {
      struct pxa_timer_freq_chout freq_chout;

      if (copy_from_user(&freq_chout, (void*) arg, sizeof(freq_chout))) {
	ret = -EFAULT;
	break;
      }
 
      /* sanity check */
      if ((freq_chout.nr >= NR_PXA_CHOUT) ||
	  (freq_chout.f_hz == 0) ||
	  (freq_chout.ratio < 1)) {
	ret = -EINVAL;
	break;
      }
      down(&pt_ctx.sem);
      chout = &(pt_ctx.chout[freq_chout.nr]);
      pxa_timer_set_freq(freq_chout.f_hz, freq_chout.ratio, chout);
      up(&pt_ctx.sem);
      break;
    }

    default:
      ret = -EINVAL;
    }

  return ret;
}


static int pxa_timer_release(struct inode *inode, struct file *filp)
{
  /* free all used GPIOs if PXAIOCTMRREM** has not been called */
  struct pxa_timer_chout *chout;
  down(&pt_ctx.sem);
  for (chout=pt_ctx.chout; chout<&pt_ctx.chout[NR_PXA_CHOUT]; chout++) {
    if (!(chout->idle) && (chout->pid == current->pid)) {
      pxa_timer_unuse_chout(chout);
    }
  }
  up(&pt_ctx.sem);

  pt_ctx.use_count--;
  return 0;
}

static struct file_operations pxa_timer_fops = {
	owner:		THIS_MODULE,
	open:		pxa_timer_open,
	release:	pxa_timer_release,
	ioctl:		pxa_timer_ioctl,
};


/*
 * This is an unofficial PXA extra timers driver considered as a miscdevice:
 *   10 char        Non-serial mice, misc features
 *                  20 = /dev/pxa-timer
 */
static struct miscdevice pxa_timer_dev = {
	minor:	PXA_TIMER_MINOR,
	name:	"pxa-timer",
	fops:	&pxa_timer_fops,
};


/*
 * Initialization.
 */
static int __init pxa_timer_init(void)
{
  init_MUTEX(&pt_ctx.sem);

  pt_ctx.chout[0].omcr = (u32*) &(OMCR10);
  pt_ctx.chout[0].osmr = (u32*) &(OSMR10);
  pt_ctx.chout[0].oscr = (u32*) &(OSCR10);
  pt_ctx.chout[1].omcr = (u32*) &(OMCR11);
  pt_ctx.chout[1].osmr = (u32*) &(OSMR11);
  pt_ctx.chout[1].oscr = (u32*) &(OSCR11);

#if !defined(PXA_TIMER_GPIO) || !defined(PXA_TIMER_GPIO_ALT_FN)
#error "TIMER GPIOs list and alternate functions must be defined"
#endif
  unsigned int gpio_list[] = PXA_TIMER_GPIO;
  unsigned int alt_fn_list[] = PXA_TIMER_GPIO_ALT_FN;
  int ii;
  for (ii=0; ii<NR_PXA_CHOUT; ii++) {
    pt_ctx.chout[ii].gpio   = gpio_list[ii];
    pt_ctx.chout[ii].alt_fn = alt_fn_list[ii];
  }

  return misc_register(&pxa_timer_dev);
}

static void __exit pxa_timer_exit(void)
{
  misc_deregister(&pxa_timer_dev);
}

module_init(pxa_timer_init);
module_exit(pxa_timer_exit);

MODULE_AUTHOR("Florian Pantaleao <fpantaleao@mobisensesystems.com>");
MODULE_DESCRIPTION("PXA extra timers driver");
MODULE_LICENSE("GPL");
