/*
 * V4L2 Driver for PXA camera host
 *
 * Copyright (C) 2006, Sascha Hauer, Pengutronix
 * Copyright (C) 2008, Guennadi Liakhovetski <kernel@pengutronix.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */


/* #define DEBUG2 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/moduleparam.h>
#include <linux/time.h>
#include <linux/version.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/mutex.h>

#include <media/v4l2-common.h>
#include <media/v4l2-dev.h>
#include <media/soc_camera.h>

#include <linux/videodev.h>
#include <linux/video_decoder.h>

#include <asm/dma.h>
#include <asm/io.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/camera.h>

#define PXA_CAM_VERSION_CODE KERNEL_VERSION(0, 0, 5)
#define PXA_CAM_DRV_NAME "pxa27x-camera"

#define CICR0_SIM_MP	(0 << 24)
#define CICR0_SIM_SP	(1 << 24)
#define CICR0_SIM_MS	(2 << 24)
#define CICR0_SIM_EP	(3 << 24)
#define CICR0_SIM_ES	(4 << 24)

#define CICR1_DW_VAL(x)   ((x) & CICR1_DW)	    /* Data bus width */
#define CICR1_PPL_VAL(x)  (((x) << 15) & CICR1_PPL) /* Pixels per line */

#define CICR2_BLW_VAL(x)  (((x) << 24) & CICR2_BLW) /* Beginning-of-line pixel clock wait count */
#define CICR2_ELW_VAL(x)  (((x) << 16) & CICR2_ELW) /* End-of-line pixel clock wait count */
#define CICR2_HSW_VAL(x)  (((x) << 10) & CICR2_HSW) /* Horizontal sync pulse width */
#define CICR2_BFPW_VAL(x) (((x) << 3) & CICR2_BFPW) /* Beginning-of-frame pixel clock wait count */
#define CICR2_FSW_VAL(x)  (((x) << 0) & CICR2_FSW)  /* Frame stabilization wait count */

#define CICR3_BFW_VAL(x)  (((x) << 24) & CICR3_BFW) /* Beginning-of-frame line clock wait count  */
#define CICR3_EFW_VAL(x)  (((x) << 16) & CICR3_EFW) /* End-of-frame line clock wait count */
#define CICR3_VSW_VAL(x)  (((x) << 11) & CICR3_VSW) /* Vertical sync pulse width */
#define CICR3_LPF_VAL(x)  (((x) << 0) & CICR3_LPF)  /* Lines per frame */

/* #define CICR0_IRQ_MASK (CICR0_TOM | CICR0_RDAVM | CICR0_FEM | CICR0_EOLM | \ */
/* 			CICR0_PERRM | CICR0_QDM | CICR0_CDM | CICR0_SOFM | \ */
/* 			CICR0_EOFM | CICR0_FOM) */
#define CICR0_IRQ_MASK (CICR0_TOM | CICR0_RDAVM | CICR0_FEM | CICR0_EOLM | \
			CICR0_PERRM | CICR0_QDM | CICR0_CDM | CICR0_FOM)

static DEFINE_MUTEX(camera_lock);

/*
 * Structures
 */

struct pxa_dmachain_attr {
	int sglen;
	unsigned int sg_offset;
	unsigned int size;
	u32 cibr;
	int endirqen;
};

/* buffer for one video frame */
struct pxa_buffer {
	/* common v4l buffer stuff -- must be first */
	struct videobuf_buffer vb;

	const struct soc_camera_data_format        *fmt;

	/* our descriptor list needed for the PXA DMA engine */
	dma_addr_t		sg_dma;
	dma_addr_t		sg_dma_cb;
	dma_addr_t		sg_dma_cr;
	struct pxa_dma_desc	*sg_cpu;
	size_t			sg_size;
	int			sgoffset_cb;
	int			sgoffset_cr;
};

struct pxa_camera_dev {
	struct device		*dev;
	/* PXA27x is only supposed to handle one camera on its Quick Capture
	 * interface. If anyone ever builds hardware to enable more than
	 * one camera, they will have to modify this driver too */
	struct soc_camera_device *icd;

	unsigned int		irq;
	void __iomem		*base;
	unsigned int		dma_chan_y;
	unsigned int		dma_chan_cb;
	unsigned int		dma_chan_cr;

	enum v4l2_buf_type	type;

	struct pxacamera_platform_data *pdata;
	struct resource		*res;
	unsigned long		platform_flags;
	unsigned long		platform_mclk_10khz;

	struct list_head	capture;

	spinlock_t		lock;

	int			start_dma;

	struct pxa_buffer	*active;

	unsigned int		frame_count;
};

static const char *pxa_cam_driver_description = "PXA_Camera";

static unsigned int vid_limit = 16;	/* Video memory limit, in Mb */

/*
 *  Videobuf operations
 */
#define PAGE_ALIGN_SIZE_Y(width,height) (PAGE_ALIGN(width * height))
#define PAGE_ALIGN_SIZE_C(width,height) (PAGE_ALIGN((width>>1) * height))

static unsigned int pxa_videobuf_size(unsigned int width, unsigned int height,
				      unsigned int depth)
{
  unsigned int size;

  if (CICR1 & CICR1_YCBCR_F)
    /* planarized format: each image plane must be page aligned */
    size = PAGE_ALIGN_SIZE_Y(width, height) +
           2 * PAGE_ALIGN_SIZE_C(width, height);
  else
    size = width * height * ((depth + 7) >> 3);

  return size;
}

static int pxa_videobuf_setup(struct videobuf_queue *vq, unsigned int *count,
			      unsigned int *size)
{
	struct soc_camera_device *icd = vq->priv_data;

	dev_dbg(&icd->dev, "count=%d, size=%d\n", *count, *size);

	*size = pxa_videobuf_size(icd->width, icd->height,
				  icd->current_fmt->depth);

	if (0 == *count)
		*count = 32;
	while (*size * *count > vid_limit * 1024 * 1024)
		(*count)--;

	return 0;
}

static void free_buffer(struct videobuf_queue *vq, struct pxa_buffer *buf)
{
	struct soc_camera_device *icd = vq->priv_data;
	struct soc_camera_host *ici =
		to_soc_camera_host(icd->dev.parent);
	struct pxa_camera_dev *pcdev = ici->priv;

	BUG_ON(in_interrupt());

	dev_dbg(pcdev->dev, "%s (vb=0x%p) 0x%08lx %d\n", __FUNCTION__,
		&buf->vb, buf->vb.baddr, buf->vb.bsize);

	/* This waits until this buffer is out of danger, i.e., until it is no
	 * longer in STATE_QUEUED or STATE_ACTIVE */
	videobuf_waiton(&buf->vb, 0, 0);
	videobuf_dma_unmap(vq, &buf->vb.dma);
	videobuf_dma_free(&buf->vb.dma);

	if (buf->sg_cpu)
		dma_free_coherent(pcdev->dev, buf->sg_size, buf->sg_cpu,
				  buf->sg_dma);
	buf->sg_cpu = NULL;

	buf->vb.state = STATE_NEEDS_INIT;
}

static void pxa_camera_dma_init(struct pxa_camera_dev *pcdev,
				struct pxa_buffer *buf,
				struct pxa_dmachain_attr *dca)
{
	struct scatterlist *sg = &buf->vb.dma.sglist[dca->sg_offset];
	struct pxa_dma_desc *sg_cpu = &buf->sg_cpu[dca->sg_offset];
	unsigned int dma_len, xfer_len, size;
	int i;

	dev_dbg(pcdev->dev, "size = %u, sglen = %d, sg_offset = %u\n",
	       dca->size, dca->sglen, dca->sg_offset);
	size = dca->size;
	for (i = 0; i < dca->sglen; i++ ) {
		BUG_ON(!sg[i].page);
		dma_len = sg_dma_len(&sg[i]);

		sg[i].dma_address = page_to_phys(sg[i].page) + sg[i].offset;

		/* CIBRx */
		sg_cpu[i].dsadr = dca->cibr;
		sg_cpu[i].dtadr = sg_dma_address(&sg[i]);
		/* PXA270 Developer's Manual 27.4.4.1:
		 * round up to 8 bytes */
		xfer_len = (min(dma_len, size) + 7) & ~7;
		if (xfer_len & 7)
			dev_err(pcdev->dev, "Unaligned buffer: "
				"dma_len %u, size %u\n", dma_len, size);
		sg_cpu[i].dcmd = DCMD_FLOWSRC | DCMD_BURST32 |
				 DCMD_INCTRGADDR | xfer_len;
		size -= dma_len;
		sg_cpu[i].ddadr = buf->sg_dma + (dca->sg_offset + i + 1) *
				sizeof(struct pxa_dma_desc);
	}
	sg_cpu[i - 1].ddadr = DDADR_STOP;
	if (dca->endirqen)
		sg_cpu[i - 1].dcmd |= DCMD_ENDIRQEN;
}

static int pxa_videobuf_prepare(struct videobuf_queue *vq,
		struct videobuf_buffer *vb, enum v4l2_field field)
{
	struct soc_camera_device *icd = vq->priv_data;
	struct soc_camera_host *ici =
		to_soc_camera_host(icd->dev.parent);
	struct pxa_camera_dev *pcdev = ici->priv;
	struct pxa_buffer *buf = container_of(vb, struct pxa_buffer, vb);
	int ret;

	dev_dbg(pcdev->dev, "%s (vb=0x%p) 0x%08lx %d\n", __FUNCTION__,
		vb, vb->baddr, vb->bsize);

#ifdef DEBUG
	/* This can be useful if you want to see if we actually fill
	 * the buffer with something */
	memset((void *)vb->baddr, 0xaa, vb->bsize);
#endif

	BUG_ON(NULL == icd->current_fmt);

	if (buf->fmt	!= icd->current_fmt ||
	    vb->width	!= icd->width ||
	    vb->height	!= icd->height ||
	    vb->field	!= field) {
		buf->fmt	= icd->current_fmt;
		vb->width	= icd->width;
		vb->height	= icd->height;
		vb->field	= field;
		vb->state	= STATE_NEEDS_INIT;
	}

	vb->size = pxa_videobuf_size(vb->width, vb->height, buf->fmt->depth);
	if (0 != vb->baddr && vb->bsize < vb->size) {
		ret = -EINVAL;
		goto out;
	}

	if (vb->state == STATE_NEEDS_INIT) {
		struct pxa_dmachain_attr dca;

		ret = videobuf_iolock(vq, vb, NULL);
		if (ret)
			goto fail;

		if (buf->sg_cpu)
			dma_free_coherent(pcdev->dev, buf->sg_size, buf->sg_cpu,
					  buf->sg_dma);

		buf->sg_size = vb->dma.sglen * sizeof(struct pxa_dma_desc);
		buf->sg_cpu = dma_alloc_coherent(pcdev->dev, buf->sg_size,
						 &buf->sg_dma, GFP_KERNEL);
		if (!buf->sg_cpu) {
			ret = -ENOMEM;
			goto fail;
		}

		dev_dbg(pcdev->dev, "nents=%d size: %lu sg=0x%p\n",
			vb->dma.sglen, vb->size, vb->dma.sglist);

		if (CICR1 & CICR1_YCBCR_F) {
			/* planarized format */
			dca.size = vb->width * vb->height;
			dca.sglen = PAGE_ALIGN(dca.size) >> PAGE_SHIFT;
			dca.sg_offset = 0;
			dca.cibr = pcdev->res->start + 0x28;
			dca.endirqen = 1;
			pxa_camera_dma_init(pcdev, buf, &dca);

			/* No IRQ at end of Cb/Cr acquisition */
			dca.sg_offset = dca.sglen;
			dca.size = dca.size >> 1;
			dca.sglen = PAGE_ALIGN(dca.size) >> PAGE_SHIFT;
			dca.cibr = pcdev->res->start + 0x30;
			dca.endirqen = 0;
			pxa_camera_dma_init(pcdev, buf, &dca);
			buf->sgoffset_cb = dca.sg_offset;
			buf->sg_dma_cb = buf->sg_dma +
				dca.sg_offset * sizeof(struct pxa_dma_desc);

			dca.sg_offset += dca.sglen;
			dca.cibr = pcdev->res->start + 0x38;
			pxa_camera_dma_init(pcdev, buf, &dca);
			buf->sgoffset_cr = dca.sg_offset;
			buf->sg_dma_cr = buf->sg_dma +
				dca.sg_offset * sizeof(struct pxa_dma_desc);
		}
		else {
			/* packed format */
			dca.sglen = vb->dma.sglen;
			dca.sg_offset = 0;
			dca.cibr = pcdev->res->start + 0x28;
			dca.endirqen = 1;
			pxa_camera_dma_init(pcdev, buf, &dca);
		}

#ifdef DEBUG2
		int j;
		struct scatterlist *_sg = vb->dma.sglist;
		struct pxa_dma_desc *sg_cpu = buf->sg_cpu;
		for (j=0; j<vb->dma.sglen; j++ )
		  printk(KERN_DEBUG "%d - dma_addr 0x%08x - sg_dma 0x%08x - dsadr 0x%08x - dtadr 0x%08x - dcmd 0x%08x - ddadr 0x%08x\n",
			 j,
			 _sg[j].dma_address,
			 buf->sg_dma + j * sizeof(struct pxa_dma_desc),
			 sg_cpu[j].dsadr,
			 sg_cpu[j].dtadr,
			 sg_cpu[j].dcmd,
			 sg_cpu[j].ddadr);
#endif			 
		vb->state = STATE_PREPARED;
	}

	return 0;

fail:
	free_buffer(vq, buf);
out:
	return ret;
}

static void pxa_camera_dma_start(struct pxa_camera_dev *pcdev)
{
	/* pcdev->active must point the first buffer to fill */
	dev_dbg(pcdev->dev, "DMA Y  start = 0x%08x\n", pcdev->active->sg_dma);

	CIFR |= CIFR_RESET_F;
	if (CICR1 & CICR1_YCBCR_F) {
		dev_dbg(pcdev->dev, "DMA Cb start = 0x%08x\n",
			pcdev->active->sg_dma_cb);
		dev_dbg(pcdev->dev, "DMA Cr start = 0x%08x\n",
			pcdev->active->sg_dma_cr);
		DDADR(pcdev->dma_chan_cb) = pcdev->active->sg_dma_cb;
		DCSR(pcdev->dma_chan_cb) = DCSR_RUN;
		DDADR(pcdev->dma_chan_cr) = pcdev->active->sg_dma_cr;
		DCSR(pcdev->dma_chan_cr) = DCSR_RUN;
	}
	DDADR(pcdev->dma_chan_y) = pcdev->active->sg_dma;
	DCSR(pcdev->dma_chan_y) = DCSR_RUN;
}

static void pxa_videobuf_queue(struct videobuf_queue *vq,
			       struct videobuf_buffer *vb)
{
	struct soc_camera_device *icd = vq->priv_data;
	struct soc_camera_host *ici =
		to_soc_camera_host(icd->dev.parent);
	struct pxa_camera_dev *pcdev = ici->priv;
	struct pxa_buffer *buf = container_of(vb, struct pxa_buffer, vb);
	struct pxa_buffer *last;
	unsigned long flags;

	dev_dbg(pcdev->dev, "%s (vb=0x%p) 0x%08lx %d\n", __FUNCTION__,
		vb, vb->baddr, vb->bsize);

	vb->state = STATE_ACTIVE;

	if (!(pcdev->active)) {
		list_add_tail(&vb->queue, &pcdev->capture);
		pcdev->active = buf;
		pcdev->start_dma = 1;
	}
	else {
		/* Add descriptors at the end of the chain.
		 * Note DMA may stop during this. That case is
		 * handled in the DMA IRQ handler. */
		spin_lock_irqsave(&pcdev->lock, flags);
		list_add_tail(&vb->queue, &pcdev->capture);
		last = list_entry(vb->queue.prev, struct pxa_buffer, vb.queue);
		if (CICR1 & CICR1_YCBCR_F) {
			/* append Cb/Cr DMA chains first, Y at last */
			last->sg_cpu[last->vb.dma.sglen - 1].ddadr =
				buf->sg_dma_cr;
			last->sg_cpu[last->sgoffset_cr - 1].ddadr =
				buf->sg_dma_cb;
			last->sg_cpu[last->sgoffset_cb - 1].ddadr =
				buf->sg_dma;
		}
		else {
			last->sg_cpu[last->vb.dma.sglen - 1].ddadr = buf->sg_dma;
		}
		spin_unlock_irqrestore(&pcdev->lock, flags);
	}
}

static void pxa_videobuf_release(struct videobuf_queue *vq,
				 struct videobuf_buffer *vb)
{
	struct pxa_buffer *buf = container_of(vb, struct pxa_buffer, vb);
#ifdef DEBUG
	struct soc_camera_device *icd = vq->priv_data;
	struct soc_camera_host *ici =
		to_soc_camera_host(icd->dev.parent);
	struct pxa_camera_dev *pcdev = ici->priv;

	dev_dbg(pcdev->dev, "%s (vb=0x%p) 0x%08lx %d\n", __FUNCTION__,
		vb, vb->baddr, vb->bsize);

	switch(vb->state) {
	case STATE_ACTIVE:
		dev_dbg(pcdev->dev, "%s (active)\n", __FUNCTION__);
		break;
	case STATE_QUEUED:
		dev_dbg(pcdev->dev, "%s (queued)\n", __FUNCTION__);
		break;
	case STATE_PREPARED:
		dev_dbg(pcdev->dev, "%s (prepared)\n", __FUNCTION__);
		break;
	case STATE_DONE:
		dev_dbg(pcdev->dev, "%s (done)\n", __FUNCTION__);
		break;
	case STATE_IDLE:
		dev_dbg(pcdev->dev, "%s (idle)\n", __FUNCTION__);
		break;
	default:
		dev_dbg(pcdev->dev, "%s (unknown)\n", __FUNCTION__);
		break;
	}
#endif

	free_buffer(vq, buf);
}

static int pxa_videobuf_map_sg(void *dev, struct scatterlist *sg, int nents,
		       int direction)
{
	pr_debug("%s, number of pages=%d\n", __FUNCTION__,nents);
	BUG_ON(direction == DMA_NONE);

	dma_map_sg(dev, sg, nents, DMA_FROM_DEVICE);

	return nents;
}

static int pxa_videobuf_unmap_sg(void *dev, struct scatterlist *sg, int nents,
			 int direction)
{
	pr_debug("%s\n", __FUNCTION__);
	dma_unmap_sg(dev, sg, nents, DMA_FROM_DEVICE);
	return 0;
}

static int pxa_videobuf_dma_sync_sg(void *dev, struct scatterlist *sg, int nents,
			    int direction)
{
	return 0;
}

static void pxa_camera_dma_irq_y(int channel, void *data)
{
	struct pxa_camera_dev *pcdev = data;
	struct pxa_buffer *buf;
	unsigned long flags;
	unsigned int status;
	struct videobuf_buffer *vb;

	spin_lock_irqsave(&pcdev->lock, flags);

	status = DCSR(pcdev->dma_chan_y);
	DCSR(pcdev->dma_chan_y) = status;

#ifdef DEBUG
	struct timespec ts;
	getnstimeofday(&ts);
	printk(KERN_DEBUG "%ld.%09ld DMA Y (%d), DCSR=0x%08x, DDADR=0x%08x\n",
	       ts.tv_sec, ts.tv_nsec, pcdev->dma_chan_y, status,
	       DDADR(pcdev->dma_chan_y));
#endif

	if (status & DCSR_BUSERR) {
		dev_err(pcdev->dev, "DMA Bus Error IRQ!\n");
		goto out;
	}

	if (!(status & DCSR_ENDINTR)) {
		dev_err(pcdev->dev, "Unknown DMA IRQ source, "
			"status: 0x%08x\n", status);
		goto out;
	}

	if (!pcdev->active) {
		dev_err(pcdev->dev, "DMA End IRQ with no active buffer!\n");
		goto out;
	}

	buf = pcdev->active;
	vb = &buf->vb;
	WARN_ON(list_empty(&vb->queue));
/* 	dev_dbg(pcdev->dev, "%s (vb=0x%p) 0x%08lx %d\n", __FUNCTION__, */
/* 		vb, vb->baddr, vb->bsize); */

	list_del(&vb->queue);
	vb->state = STATE_DONE;
	do_gettimeofday(&vb->ts);
	vb->field_count = pcdev->frame_count;
	/* restore DDADR_STOP at end of chain that might have been deleted
	 * in pxa_videobuf_queue */
	buf->sg_cpu[vb->dma.sglen - 1].ddadr = DDADR_STOP;
	if (CICR1 & CICR1_YCBCR_F) {
		buf->sg_cpu[buf->sgoffset_cr - 1].ddadr = DDADR_STOP;
		buf->sg_cpu[buf->sgoffset_cb - 1].ddadr = DDADR_STOP;
	}
	wake_up(&vb->done);

	if (list_empty(&pcdev->capture)) {
		pcdev->active = NULL;
		if (DCSR(pcdev->dma_chan_y) & DCSR_RUN) {
		  DCSR(pcdev->dma_chan_y) &= ~DCSR_RUN;
		  dev_warn(pcdev->dev,
			   "   !!! capture empty, DMA not stopped !!!\n");
		}
		DCSR(pcdev->dma_chan_cb) &= ~DCSR_RUN;
		DCSR(pcdev->dma_chan_cr) &= ~DCSR_RUN;
		goto out;
	}

	pcdev->active = list_entry(pcdev->capture.next, struct pxa_buffer,
				   vb.queue);

	if (!(DCSR(pcdev->dma_chan_y) & DCSR_RUN)) {
		/* exceptionnal situation where DMA stopped while appending
		 * a descriptor chain => restart DMA */
		pcdev->start_dma = 1;
	}

out:
	spin_unlock_irqrestore(&pcdev->lock, flags);
}

static void pxa_camera_dma_irq_cb(int channel, void *data)
{
	/* dummy handler, never called except fro debug */
	struct pxa_camera_dev *pcdev = data;
	unsigned int status;
	status = DCSR(pcdev->dma_chan_cb);
	DCSR(pcdev->dma_chan_cb) = status;

	struct timespec ts;
	getnstimeofday(&ts);
	printk(KERN_DEBUG "%ld.%09ld DMA Cb(%d), DCSR=0x%08x, DDADR=0x%08x\n",
	       ts.tv_sec, ts.tv_nsec, pcdev->dma_chan_cb, status,
	       DDADR(pcdev->dma_chan_cb));
}

static void pxa_camera_dma_irq_cr(int channel, void *data)
{
	/* dummy handler, never called except fro debug */
	struct pxa_camera_dev *pcdev = data;
	unsigned int status;
	status = DCSR(pcdev->dma_chan_cr);
	DCSR(pcdev->dma_chan_cr) = status;

	struct timespec ts;
	getnstimeofday(&ts);
	printk(KERN_DEBUG "%ld.%09ld DMA Cr(%d), DCSR=0x%08x, DDADR=0x%08x\n",
	       ts.tv_sec, ts.tv_nsec, pcdev->dma_chan_cr, status,
	       DDADR(pcdev->dma_chan_cr));
}

static struct videobuf_queue_ops pxa_videobuf_ops = {
	.buf_setup      = pxa_videobuf_setup,
	.buf_prepare    = pxa_videobuf_prepare,
	.buf_queue      = pxa_videobuf_queue,
	.buf_release    = pxa_videobuf_release,

	/* Non-pci handling routines */
	.vb_map_sg      = pxa_videobuf_map_sg,
	.vb_dma_sync_sg = pxa_videobuf_dma_sync_sg,
	.vb_unmap_sg    = pxa_videobuf_unmap_sg,
};

static int mclk_get_divisor(unsigned int mclk_10khz)
{
	unsigned long div;
	unsigned long lcdclk;

	lcdclk = get_lcdclk_frequency_10khz();

	/* We verify platform_mclk_10khz != 0, so if anyone breaks it, here
	 * they get a nice Oops */
	div = (lcdclk + 2 * mclk_10khz - 1) / (2 * mclk_10khz) - 1;

	return div;
}

static void pxa_camera_setup(struct pxa_camera_dev *pcdev)
{
	struct pxacamera_platform_data *pdata = pcdev->pdata;
	u32 cicr4 = 0;
	dev_dbg(pcdev->dev, "pxa_camera_setup begin\n");

	CICR0 = 0x3FF;   /* disable all interrupts */

	if (pcdev->platform_flags & PXA_CAMERA_PCLK_EN)
		cicr4 |= CICR4_PCLK_EN;
	if (pcdev->platform_flags & PXA_CAMERA_MCLK_EN)
		cicr4 |= CICR4_MCLK_EN;
	if (pcdev->platform_flags & PXA_CAMERA_PCP)
		cicr4 |= CICR4_PCP;
	if (pcdev->platform_flags & PXA_CAMERA_HSP)
		cicr4 |= CICR4_HSP;
	if (pcdev->platform_flags & PXA_CAMERA_VSP)
		cicr4 |= CICR4_VSP;

	CICR4 = mclk_get_divisor(pcdev->platform_mclk_10khz) | cicr4;
	CISR  = ~0;
	CIFR  =  0;
	CITOR =  0;

	if (pcdev->platform_flags & PXA_CAMERA_MCLK_EN)
		CICR0 |= CICR0_ENB;	/* keep CI on for MCLK */

	dev_dbg(pcdev->dev, "CICR0 0x%08x, CICR4 0x%08x\n", CICR0, CICR4);
	dev_dbg(pcdev->dev, "pxa_camera_setup end\n");
}

static int pxa_camera_enable(struct soc_camera_device *icd)
{
	struct soc_camera_host *ici = to_soc_camera_host(icd->dev.parent);
	struct pxa_camera_dev *pcdev = ici->priv;
	struct pxacamera_platform_data *pdata = pcdev->pdata;
	int ret;
	dev_dbg(&icd->dev, "pxa_camera_enable. ici 0x%p, icd 0x%p, pcdev 0x%p\n",
		ici, icd, pcdev);

	pcdev->start_dma = 0;
	pcdev->active = NULL;

	pxa_set_cken(CKEN24_CAMERA, 1);

	/* TODO should be in sensor driver code */
	if (pdata && pdata->power) {
		dev_dbg(pcdev->dev, "%s: Power on camera\n", __FUNCTION__);
		pdata->power(pcdev->dev, 1);
	}

	if (pdata && pdata->reset) {
		dev_dbg(pcdev->dev, "%s: Releasing camera reset\n",
			__FUNCTION__);
		pdata->reset(pcdev->dev, 1);
	}
	/* END TODO */

	ret = icd->ops->init(icd,
			(pcdev->platform_flags & PXA_CAMERA_MASTER ?
			 IS_MASTER : 0) |
			(pcdev->platform_flags & PXA_CAMERA_HSP ?
			 0 : IS_HSYNC_ACTIVE_HIGH) |
			(pcdev->platform_flags & PXA_CAMERA_VSP ?
			 0 : IS_VSYNC_ACTIVE_HIGH) |
			(pcdev->platform_flags & PXA_CAMERA_PCP ?
			 0 : IS_PCLK_SAMPLE_RISING));
	if (ret)
		pxa_set_cken(CKEN24_CAMERA, 0);

	return ret;
}

static int pxa_camera_disable(struct soc_camera_device *icd)
{
	struct soc_camera_host *ici = to_soc_camera_host(icd->dev.parent);
	struct pxa_camera_dev *pcdev = ici->priv;
	struct pxacamera_platform_data *board = pcdev->pdata;

	/* disable capture, disable interrupts */
	CICR0 = 0x3ff;
	if (pcdev->platform_flags & PXA_CAMERA_MCLK_EN)
		CICR0 |= CICR0_ENB;	/* keep CI on for MCLK */

	/* Stop DMA engine */
	DCSR(pcdev->dma_chan_y ) = 0;
	DCSR(pcdev->dma_chan_cb) = 0;
	DCSR(pcdev->dma_chan_cr) = 0;

	icd->ops->release(icd);

	if (!(pcdev->platform_flags & PXA_CAMERA_MCLK_EN))
		pxa_set_cken(CKEN24_CAMERA, 0);

	/* TODO should be in sensor driver code */
	if (board && board->reset) {
		dev_dbg(pcdev->dev, "%s: Asserting camera reset\n",
			__FUNCTION__);
		board->reset(pcdev->dev, 0);
	}

	if (board && board->power) {
		dev_dbg(pcdev->dev, "%s: Power off camera\n", __FUNCTION__);
		board->power(pcdev->dev, 0);
	}
	/* END TODO */

	return 0;
}

static irqreturn_t pxa_camera_irq(int irq, void *data)
{
	struct pxa_camera_dev *pcdev = data;
	unsigned int status = CISR;
#ifdef DEBUG
	struct timespec ts;
	getnstimeofday(&ts);
#endif

	if (status & CISR_SOF) {
		if (pcdev->start_dma) {
			pxa_camera_dma_start(pcdev);
			pcdev->start_dma = 0;
		}
		/* increment frame_count by 2 because we do progressive scan */
		pcdev->frame_count += 2;
#ifdef DEBUG
		if (pcdev->active) {
			dev_dbg(pcdev->dev, "%ld.%09ld SOF %d\n",
				ts.tv_sec, ts.tv_nsec, pcdev->frame_count);
/* 	    		CICR0 &= ~CICR0_EOLM; */
			CICR0 &= ~CICR0_EOFM;
		}
	}
	else if (status & CISR_EOF) {
		dev_dbg(pcdev->dev, "%ld.%09ld EOF\n", ts.tv_sec, ts.tv_nsec);
		CICR0 |= CICR0_EOFM;
	}
	else if (status & CISR_EOL) {
		dev_dbg(pcdev->dev, "%ld.%09ld EOL 1\n", ts.tv_sec, ts.tv_nsec);
		CICR0 |= CICR0_EOLM;
	}
#else
	}
#endif
	else if (status & CISR_IFO_0)
		dev_dbg(pcdev->dev, "F0O\n");
	else if (status & CISR_IFO_1)
		dev_dbg(pcdev->dev, "F1O\n");
	else if (status & CISR_IFO_2)
		dev_dbg(pcdev->dev, "F2O\n");
	else
		dev_dbg(pcdev->dev, "Camera interrupt status 0x%x\n", status);

	CISR = status;

	return IRQ_HANDLED;
}

/* The following two functions absolutely depend on the fact, that
 * there can be only one camera on PXA quick capture interface */
static int pxa_camera_add_device(struct soc_camera_device *icd)
{
	struct soc_camera_host *ici = to_soc_camera_host(icd->dev.parent);
	struct pxa_camera_dev *pcdev = ici->priv;
	int ret = 0;

	mutex_lock(&camera_lock);

	if (pcdev->icd) {
		ret = -EBUSY;
		goto ebusy;
	}

	dev_info(pcdev->dev, "PXA Camera driver attached to camera %d\n",
		 icd->devnum);
	dev_dbg(&icd->dev, "pxa_camera_add_device. ici 0x%p, icd 0x%p, pcdev 0x%p\n",
		ici, icd, pcdev);

	pxa_camera_setup(pcdev);

	pcdev->icd = icd;

ebusy:
	mutex_unlock(&camera_lock);

	return ret;
}

static void pxa_camera_remove_device(struct soc_camera_device *icd)
{
	struct soc_camera_host *ici = to_soc_camera_host(icd->dev.parent);
	struct pxa_camera_dev *pcdev = ici->priv;

	BUG_ON(icd != pcdev->icd);

	dev_info(pcdev->dev, "PXA Camera driver detached from camera %d\n",
		 icd->devnum);

	pcdev->icd = NULL;
}

static int pxa_camera_set_capture_format(struct soc_camera_device *icd,
					 __u32 pixfmt, struct v4l2_rect *rect)
{
	struct soc_camera_host *ici =
		to_soc_camera_host(icd->dev.parent);
	struct pxa_camera_dev *pcdev = ici->priv;
	unsigned int datawidth = 0, dw, color_sp, ycbcr_f, raw_bpp;
	u32 cicr0, cicr4 = 0;
	int ret;

	/* If requested data width is supported by the platform, use it */
	switch (icd->cached_datawidth) {
	case 10:
		if (pcdev->platform_flags & PXA_CAMERA_DATAWIDTH_10)
			datawidth = IS_DATAWIDTH_10;
		break;
	case 9:
		if (pcdev->platform_flags & PXA_CAMERA_DATAWIDTH_9)
			datawidth = IS_DATAWIDTH_9;
		break;
	case 8:
		if (pcdev->platform_flags & PXA_CAMERA_DATAWIDTH_8)
			datawidth = IS_DATAWIDTH_8;
	}
	if (!datawidth)
		return -EINVAL;

	ret = icd->ops->set_capture_format(icd, pixfmt, rect,
			datawidth |
			(pcdev->platform_flags & PXA_CAMERA_MASTER ?
			 IS_MASTER : 0) |
			(pcdev->platform_flags & PXA_CAMERA_HSP ?
			 0 : IS_HSYNC_ACTIVE_HIGH) |
			(pcdev->platform_flags & PXA_CAMERA_VSP ?
			 0 : IS_VSYNC_ACTIVE_HIGH) |
			(pcdev->platform_flags & PXA_CAMERA_PCP ?
			 0 : IS_PCLK_SAMPLE_RISING));
	if (ret < 0)
		return ret;

	/* Datawidth is now guaranteed to be equal to one of the three values.
	 * We fix bit-per-pixel equal to data-width... */
	switch (datawidth) {
	case IS_DATAWIDTH_10:
		icd->cached_datawidth = 10;
		dw = 4;
		color_sp = 0;
		raw_bpp = 2;
		ycbcr_f = 0;
		break;
	case IS_DATAWIDTH_9:
		icd->cached_datawidth = 9;
		dw = 3;
		color_sp = 0;
		raw_bpp = 1;
		ycbcr_f = 0;
		break;
	default:
		/* Actually it can only be 8 now,
		 * default is just to silence compiler warnings */
	case IS_DATAWIDTH_8:
		icd->cached_datawidth = 8;
		dw = 2;
		raw_bpp = 0;
		ycbcr_f = 0;
		if (pixfmt == V4L2_PIX_FMT_GREY)
			 color_sp = 0;
		else if ((pixfmt == V4L2_PIX_FMT_YUYV) ||
			 (pixfmt == V4L2_PIX_FMT_UYVY) ||
			 (pixfmt == V4L2_PIX_FMT_SBGGR8))
			 color_sp = 2;
		else if ((pixfmt == V4L2_PIX_FMT_YUV422P) ||
			 (pixfmt == V4L2_PIX_FMT_GRB422P)) {
	        	 color_sp = 2;
			 ycbcr_f = 1;
		}
	}

	if (pcdev->platform_flags & PXA_CAMERA_PCLK_EN)
		cicr4 |= CICR4_PCLK_EN;
	if (pcdev->platform_flags & PXA_CAMERA_MCLK_EN)
		cicr4 |= CICR4_MCLK_EN;
	if (pcdev->platform_flags & PXA_CAMERA_PCP)
		cicr4 |= CICR4_PCP;
	if (pcdev->platform_flags & PXA_CAMERA_HSP)
		cicr4 |= CICR4_HSP;
	if (pcdev->platform_flags & PXA_CAMERA_VSP)
		cicr4 |= CICR4_VSP;

	/* Stop interface to set new format */
	CICR0 &= ~CICR0_ENB;
	CICR1 = CICR1_PPL_VAL(rect->width - 1) |
		ycbcr_f  << 10 |
		raw_bpp  <<  5 |
		color_sp <<  3 |
		dw;
	CICR2 = 0;
	CICR3 = CICR3_LPF_VAL(rect->height - 1) |
		CICR3_BFW_VAL(min((unsigned short)255, icd->y_skip_top));
	CICR4 = mclk_get_divisor(pcdev->platform_mclk_10khz) | cicr4;

	/* set up FIFOS */
	CIFR |= CIFR_FEN0;
	if (ycbcr_f) {
		CIFR |= CIFR_FEN1;
		CIFR |= CIFR_FEN2;
	}
	else {
		CIFR &= ~CIFR_FEN1;
		CIFR &= ~CIFR_FEN2;
	}

	/* CIF and DMA interrupts are used for correct frame count */
	CICR0 = (pcdev->platform_flags & PXA_CAMERA_MASTER ?
		 CICR0_SIM_MP : (CICR0_SL_CAP_EN | CICR0_SIM_SP)) |
	  CICR0_DMAEN | CICR0_IRQ_MASK | CICR0_ENB;

	dev_dbg(pcdev->dev, "End of pxa_camera_set_capture_format:\n");
	dev_dbg(pcdev->dev,
		"CICR0 0x%08x, CICR1 0x%08x, CICR2 0x%08x, CICR3 0x%08x, CICR4 0x%08x, CIFR 0x%08x, CISR 0x%08x\n",
		CICR0, CICR1, CICR2, CICR3, CICR4, CIFR, CISR);

	return 0;
}

static int pxa_camera_try_fmt_cap(struct soc_camera_host *ici,
				  struct v4l2_format *f)
{
	/* limit to pxa hardware capabilities */
	if (f->fmt.pix.height < 32)
		f->fmt.pix.height = 32;
	if (f->fmt.pix.height > 2048)
		f->fmt.pix.height = 2048;
	if (f->fmt.pix.width < 48)
		f->fmt.pix.width = 48;
	if (f->fmt.pix.width > 2048)
		f->fmt.pix.width = 2048;
	f->fmt.pix.width &= ~0x01;

	return 0;
}

static int pxa_camera_reqbufs(struct soc_camera_file *icf,
			      struct v4l2_requestbuffers *p)
{
	int i;

	/* This is for locking debugging only. I removed spinlocks and now I
	 * check whether .prepare is ever called on a linked buffer, or whether
	 * a dma IRQ can occur for an in-work or unlinked buffer. Until now
	 * it hadn't triggered */
	for (i = 0; i < p->count; i++) {
		struct pxa_buffer *buf = container_of(icf->vb_vidq.bufs[i],
						      struct pxa_buffer, vb);
		INIT_LIST_HEAD(&buf->vb.queue);
	}

	return 0;
}

static unsigned int pxa_camera_poll(struct file *file, poll_table *pt)
{
	struct soc_camera_file *icf = file->private_data;
	struct pxa_buffer *buf;

	buf = list_entry(icf->vb_vidq.stream.next, struct pxa_buffer,
			 vb.stream);

	poll_wait(file, &buf->vb.done, pt);

	if (buf->vb.state == STATE_DONE ||
	    buf->vb.state == STATE_ERROR)
		return POLLIN|POLLRDNORM;

	return 0;
}

static int pxa_camera_querycap(struct soc_camera_host *ici,
			       struct v4l2_capability *cap)
{
	/* cap->name is set by the firendly caller:-> */
	strlcpy(cap->card, pxa_cam_driver_description, sizeof(cap->card));
	cap->version = PXA_CAM_VERSION_CODE;
	cap->capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_STREAMING;

	return 0;
}

/* Should beallocated dynamically too, but we have only one. */
static struct soc_camera_host pxa_soc_camera_host = {
	.drv_name		= PXA_CAM_DRV_NAME,
	.vbq_ops		= &pxa_videobuf_ops,
	.add			= pxa_camera_add_device,
	.remove			= pxa_camera_remove_device,
	.enable                 = pxa_camera_enable,
	.disable                = pxa_camera_disable,
	.msize			= sizeof(struct pxa_buffer),
	.set_capture_format	= pxa_camera_set_capture_format,
	.try_fmt_cap		= pxa_camera_try_fmt_cap,
	.reqbufs		= pxa_camera_reqbufs,
	.poll			= pxa_camera_poll,
	.querycap		= pxa_camera_querycap,
};

static int pxa_camera_probe(struct platform_device *pdev)
{
	struct pxa_camera_dev *pcdev;
	struct resource *res;
	void __iomem *base;
	unsigned int irq;
	int err = 0;
	dev_dbg(&pdev->dev, "pxa_camera_probe begin\n");

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	irq = platform_get_irq(pdev, 0);
	if (!res || !irq) {
		err = -ENODEV;
		goto exit;
	}

	pcdev = kzalloc(sizeof(*pcdev), GFP_KERNEL);
	if (!pcdev) {
		dev_err(&pdev->dev, "Could not allocate pcdev\n");
		err = -ENOMEM;
		goto exit;
	}

	dev_set_drvdata(&pdev->dev, pcdev);
	pcdev->res = res;
	pcdev->pdata = pdev->dev.platform_data;
	pcdev->platform_flags = pcdev->pdata->flags;
	if (!pcdev->platform_flags & (PXA_CAMERA_DATAWIDTH_8 |
			PXA_CAMERA_DATAWIDTH_9 | PXA_CAMERA_DATAWIDTH_10)) {
		/* Platform hasn't set available data widths. This is bad.
		 * Warn and use a default. */
		dev_warn(&pdev->dev, "WARNING! Platform hasn't set available "
			 "data widths, using default 10 bit\n");
		pcdev->platform_flags |= PXA_CAMERA_DATAWIDTH_10;
	}

	pcdev->platform_mclk_10khz = pcdev->pdata->mclk_10khz;
	if (!pcdev->platform_mclk_10khz) {
		dev_warn(&pdev->dev,
			 "mclk_10khz == 0! Please, fix your platform data. "
			 "Using default 20MHz\n");
		pcdev->platform_mclk_10khz = 2000;
	}
	/* set-up GPIOs */
	if (pcdev->pdata->init) {
		dev_dbg(&pdev->dev, "%s: Init gpios\n", __FUNCTION__);
		pcdev->pdata->init(&pdev->dev);
	}

	INIT_LIST_HEAD(&pcdev->capture);
	spin_lock_init(&pcdev->lock);

	/*
	 * Request the regions.
	 */
	if (!request_mem_region(res->start, res->end - res->start + 1,
				PXA_CAM_DRV_NAME)) {
		err = -EBUSY;
		goto exit_kfree;
	}
	base = ioremap(res->start, res->end - res->start + 1);
	if (!base) {
		err = -ENOMEM;
		goto exit_release;
	}
	pcdev->irq = irq;
	pcdev->base = base;
	pcdev->dev = &pdev->dev;

	/* request dma */
	pcdev->dma_chan_y = pxa_request_dma("CI_Y", DMA_PRIO_HIGH,
					    pxa_camera_dma_irq_y, pcdev);
	if (pcdev->dma_chan_y < 0) {
		dev_err(pcdev->dev, "Can't request DMA for Y\n");
		err = -ENOMEM;
		goto exit_iounmap;
	}
	DRCMR68 = pcdev->dma_chan_y  | DRCMR_MAPVLD;

	pcdev->dma_chan_cb = pxa_request_dma("CI_Cb", DMA_PRIO_HIGH,
					    pxa_camera_dma_irq_cb, pcdev);
	if (pcdev->dma_chan_cb < 0) {
		dev_err(pcdev->dev, "Can't request DMA for Cb\n");
		err = -ENOMEM;
		goto exit_free_dma_y;
	}
	DRCMR69 = pcdev->dma_chan_cb | DRCMR_MAPVLD;

	pcdev->dma_chan_cr = pxa_request_dma("CI_Cr", DMA_PRIO_HIGH,
					    pxa_camera_dma_irq_cr, pcdev);
	if (pcdev->dma_chan_cr < 0) {
		dev_err(pcdev->dev, "Can't request DMA for Cr\n");
		err = -ENOMEM;
		goto exit_free_dma_cb;
	}
	DRCMR70 = pcdev->dma_chan_cr | DRCMR_MAPVLD;

	/* request irq */
	err = request_irq(pcdev->irq, pxa_camera_irq, 0, PXA_CAM_DRV_NAME,
			  pcdev);
	if (err) {
		dev_err(pcdev->dev, "Camera interrupt register failed \n");
		goto exit_free_dma_cr;
	}

	pxa_soc_camera_host.priv	= pcdev;
	pxa_soc_camera_host.dev.parent	= &pdev->dev;
	pxa_soc_camera_host.nr		= pdev->id;
	err = soc_camera_host_register(&pxa_soc_camera_host, THIS_MODULE);

	if (err)
		goto exit_free_irq;

	pcdev->frame_count = 0;

	if (pcdev->platform_flags & PXA_CAMERA_MCLK_EN) {
		/* enable minimal operation to provide Master Clock */
		pxa_set_cken(CKEN24_CAMERA, 1);
		CICR4 = CICR4_MCLK_EN |
		  mclk_get_divisor(pcdev->platform_mclk_10khz);
		CICR0 = CICR0_ENB | 0x3ff;
		dev_dbg(&pdev->dev, "pxa_camera_probe MCLK set\n");
		dev_dbg(&pdev->dev, "CICR0=0x%08x\n", CICR0);
		dev_dbg(&pdev->dev, "CICR4=0x%08x\n", CICR4);
	}

	dev_dbg(&pdev->dev, "pxa_camera_probe end\n");

	return 0;

exit_free_irq:
	free_irq(pcdev->irq, pcdev);
exit_free_dma_cr:
	pxa_free_dma(pcdev->dma_chan_cr);
exit_free_dma_cb:
	pxa_free_dma(pcdev->dma_chan_cb);
exit_free_dma_y:
	pxa_free_dma(pcdev->dma_chan_y);
exit_iounmap:
	iounmap(base);
exit_release:
	release_mem_region(res->start, res->end - res->start + 1);
exit_kfree:
	kfree(pcdev);
exit:
	return err;
}

static int __devexit pxa_camera_remove(struct platform_device *pdev)
{
	struct pxa_camera_dev *pcdev = platform_get_drvdata(pdev);
	struct resource *res;
	dev_dbg(&pdev->dev, "pxa_camera_remove begin\n");

	if (pcdev->platform_flags & PXA_CAMERA_MCLK_EN) {
		/* disable minimal operation */
		CICR4 = 0;
		CICR0 = 0x3ff;
		pxa_set_cken(CKEN24_CAMERA, 0);
	}

	pxa_free_dma(pcdev->dma_chan_y);
	pxa_free_dma(pcdev->dma_chan_cb);
	pxa_free_dma(pcdev->dma_chan_cr);
	free_irq(pcdev->irq, pcdev);

	soc_camera_host_unregister(&pxa_soc_camera_host);

	iounmap(pcdev->base);

	res = pcdev->res;
	release_mem_region(res->start, res->end - res->start + 1);

	kfree(pcdev);

	dev_info(&pdev->dev, "PXA Camera driver unloaded\n");

	dev_dbg(&pdev->dev, "pxa_camera_remove end\n");
	return 0;
}

/*
 * Suspend the Camera Module.
 */
static int pxa_camera_suspend(struct platform_device *pdev, pm_message_t level)
{
	struct pxa_camera_dev *pcdev = platform_get_drvdata(pdev);

	dev_info(&pdev->dev, "camera suspend\n");
	disable_irq(pcdev->irq);
	return 0;
}

/*
 * Resume the Camera Module.
 */
static int pxa_camera_resume(struct platform_device *pdev)
{
	struct pxa_camera_dev *pcdev = platform_get_drvdata(pdev);

	dev_info(&pdev->dev, "camera resume\n");
	enable_irq(pcdev->irq);

	DRCMR68 = pcdev->dma_chan_y  | DRCMR_MAPVLD;
	DRCMR69 = pcdev->dma_chan_cb | DRCMR_MAPVLD;
	DRCMR70 = pcdev->dma_chan_cr | DRCMR_MAPVLD;

	return 0;
}


static struct platform_driver pxa_camera_driver = {
	.driver 	= {
		.name	= PXA_CAM_DRV_NAME,
	},
	.probe		= pxa_camera_probe,
	.remove		= __exit_p(pxa_camera_remove),
	.suspend	= pxa_camera_suspend,
	.resume		= pxa_camera_resume,
};


static int __devinit pxa_camera_init(void)
{
	return platform_driver_register(&pxa_camera_driver);
}

static void __exit pxa_camera_exit(void)
{
	return platform_driver_unregister(&pxa_camera_driver);
}

module_init(pxa_camera_init);
module_exit(pxa_camera_exit);

MODULE_DESCRIPTION("PXA27x SoC Camera Host driver");
MODULE_AUTHOR("Guennadi Liakhovetski <kernel@pengutronix.de>");
MODULE_LICENSE("GPL");
