/*
 * Omnivision OV7620 Driver with soc_camera bus
 *
 * Copyright (C) 2008, Florian Pantaleao <fpantaleao@mobisensesystems.com>
 *
 * Inspired from Guennadi Liakhovetski mt9v022.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* #define DEBUG */

#include <linux/videodev2.h>
#include <linux/ov7620.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/log2.h>
#include <linux/platform_device.h>

#include <media/v4l2-common.h>
#include <media/v4l2-chip-ident.h>
#include <media/soc_camera.h>
#include <media/soc_ov7620.h>

/* OV7620 i2c address 0x21 */

/* static char *sensor_type; */
/* module_param(sensor_type, charp, S_IRUGO); */
/* MODULE_PARM_DESC(sensor_type, "Sensor type: \"colour\" or \"monochrome\"\n"); */

static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short normal_addr[] = {I2C_CLIENT_END, I2C_CLIENT_END };

static struct i2c_client_address_data addr_data = {
	.normal_i2c	= normal_addr,
	.probe		= ignore,
	.ignore		= ignore,
};

/* Forward declarations */
static struct i2c_driver ov7620_i2c_driver;

/* Manufacture ID constants */
#define OV7620_MIDH_VAL		0x7F
#define OV7620_MIDL_VAL		0xA2

#define OV7620_COMA2            0x04
#define OV7620_COMA3            0x08
#define OV7620_COMA4            0x10
#define OV7620_COMA5            0x20
#define OV7620_COMA7            0x80
#define OV7620_COMB0            0x01
#define OV7620_COMB5            0x20
#define OV7620_COMC3            0x80
#define OV7620_COMC5            0x20
#define OV7620_COMD0            0x01
#define OV7620_COMD6            0x40
#define OV7620_COMG4            0x10
#define OV7620_COMH5            0x20
#define OV7620_COMI6            0x40
#define OV7620_COMI7            0x80
#define OV7620_COML6            0x40
#define OV7620_COMO5            0x20
#define OV7620_EHSH4            0x10
#define OV7620_FD10             0x03
#define OV7620_YUV2             0x04
#define OV7620_YUV3             0x08
#define OV7620_SPCB7            0x80
#define OV7620_SYN6             0x40

#define OV7620_SHP_MSK          0x0f

/* OV7620 selected register addresses */
#define OV7620_AGC		0x0000
#define OV7620_BLUE		0x0001
#define OV7620_RED		0x0002
#define OV7620_SAT		0x0003
#define OV7620_BRT		0x0006
#define OV7620_SHP		0x0007
#define OV7620_ABLU		0x000C
#define OV7620_ARED		0x000D
#define OV7620_AEC		0x0010
#define OV7620_CLK		0x0011
#define OV7620_COMA		0x0012
#define OV7620_COMB		0x0013
#define OV7620_COMC		0x0014
#define OV7620_COMD		0x0015
#define OV7620_FD		0x0016
#define OV7620_HSTART		0x0017
#define OV7620_HSTOP		0x0018
#define OV7620_VSTART		0x0019
#define OV7620_VSTOP		0x001A
#define OV7620_PSHFT		0x001B
#define OV7620_MIDH		0x001C		// Manufacture ID High byte
#define OV7620_MIDL		0x001D		// Manufacture ID Low byte
#define OV7620_COME		0x0020
#define OV7620_Y		0x0021
#define OV7620_U		0x0022
#define OV7620_CC		0x0023
#define OV7620_AEW		0x0024
#define OV7620_AEB		0x0025
#define OV7620_COMF		0x0026
#define OV7620_COMG		0x0027
#define OV7620_COMH		0x0028
#define OV7620_COMI		0x0029
#define OV7620_EHSH		0x002A
#define OV7620_EHSL		0x002B
#define OV7620_EXBK		0x002C
#define OV7620_COMJ		0x002D
#define OV7620_V		0x002E
#define OV7620_SPCA		0x0060
#define OV7620_SPCB		0x0061
#define OV7620_RGM		0x0062
#define OV7620_YGM		0x0064
#define OV7620_SPCC		0x0065
#define OV7620_AWBC		0x0066
#define OV7620_YUV		0x0067
#define OV7620_SPCD		0x0068
#define OV7620_EDGE		0x0069
#define OV7620_VEG		0x006A
#define OV7620_EOC		0x006F
#define OV7620_COMK		0x0070
#define OV7620_COML		0x0071
#define OV7620_HSDY		0x0072
#define OV7620_HEDY		0x0073
#define OV7620_COMM		0x0074
#define OV7620_COMN		0x0075
#define OV7620_COMO		0x0076
#define OV7620_AVG		0x007C

// End of OV7620 register
#define OV7620_REGEND		( OV7620_AVG + 1 )

/* Progressive scan, master, defaults */
/* #define MT9V022_CHIP_CONTROL_DEFAULT	0x188 */

/* OV7620 window parameters */
#define OV7620_HW_MIN  0x2C  /* registers 0x17 & 0x18 min value */
#define OV7620_HW_MAX  0xD2  /* registers 0x17 & 0x18 max value */
#define OV7620_HWS_DEF 0x2F  /* register 0x17 default value */
#define OV7620_HWE_DEF 0xCF  /* register 0x18 default value */
#define OV7620_VW_MIN  0x05  /* registers 0x19 & 0x1A min value */
#define OV7620_VW_MAX  0xF6  /* registers 0x19 & 0x1A max value */
#define OV7620_VWS_DEF 0x06  /* register 0x19 default value */
#define OV7620_VWE_DEF 0xF5  /* register 0x1A default value */
#define OV7620_W_MIN   4
#define OV7620_W_MAX   ((OV7620_HW_MAX-OV7620_HW_MIN)*4)
#define OV7620_W_DEF   ((OV7620_HWE_DEF-OV7620_HWS_DEF)*4)
#define OV7620_H_MIN   2
#define OV7620_H_MAX   ((OV7620_VW_MAX-OV7620_VW_MIN+1)*2)
#define OV7620_H_DEF   ((OV7620_VWE_DEF-OV7620_VWS_DEF+1)*2)


static const struct soc_camera_data_format ov7620_data_formats[] = {
	{
		.name		= "YUV422 planarized",
		.depth		= 16,
		.fourcc		= V4L2_PIX_FMT_YUV422P,
		.colorspace	= V4L2_COLORSPACE_JPEG,
	}, {
		.name		= "YUV422 packed UYVY",
		.depth		= 16,
		.fourcc		= V4L2_PIX_FMT_UYVY,
		.colorspace	= V4L2_COLORSPACE_JPEG,
/* 	}, { */
/* 		.name		= "YUV422 packed YUYV", */
/* 		.depth		= 16, */
/* 		.fourcc		= V4L2_PIX_FMT_YUYV, */
/* 		.colorspace	= V4L2_COLORSPACE_JPEG, */
	}, {
		.name		= "Monochrome 8 bit",
		.depth		= 8,
		.fourcc		= V4L2_PIX_FMT_GREY,
	}, {
		.name		= "Bayer planarized 8 bit",
		.depth		= 16,
		.fourcc		= V4L2_PIX_FMT_GRB422P,
		.colorspace	= V4L2_COLORSPACE_SRGB,
	}, {
		.name		= "Bayer packed 8 bit",
		.depth		= 16,
		.fourcc		= V4L2_PIX_FMT_SBGGR8,
		.colorspace	= V4L2_COLORSPACE_SRGB,
	}
};

struct ov7620 {
	struct i2c_client *client;
	struct soc_camera_device icd;
	struct soc_camera_link *icl;
	int model;	/* V4L2_IDENT_OV7620 code from v4l2-chip-ident.h */
	unsigned char datawidth;
};

struct ov7620 ov7620_ctx;

static int reg_read(struct soc_camera_device *icd, const u8 reg)
{
	struct ov7620 *ov7620 = container_of(icd, struct ov7620, icd);
	struct i2c_client *client = ov7620->client;
	char buf[1];

	/* !!! SMBUS is not supported by OV7620 !!! */
	buf[0] = reg;
        if (i2c_master_send(client, buf, 1) != 1) {
                dev_err(&icd->dev, "i2c_master_send failed.\n");
                return -1;
        }
        if (i2c_master_recv(client, buf, 1) != 1) {
                dev_err(&icd->dev, "i2c_master_recv failed.");
                return -1;
        }

	dev_dbg(&icd->dev, "reg_read 0x%02x = 0x%02x\n", reg, buf[0]);

        return buf[0];
}

static int reg_write(struct soc_camera_device *icd, const u8 reg,
		     const u8 data)
{
	dev_dbg(&icd->dev, "reg_write 0x%02x = 0x%02x\n", reg, data);
	struct ov7620 *ov7620 = container_of(icd, struct ov7620, icd);
	return i2c_smbus_write_byte_data(ov7620->client, reg, data);
}


static int reg_set(struct soc_camera_device *icd, const u8 reg,
		   const u8 data)
{
	int ret;

	ret = reg_read(icd, reg);
	if (ret < 0)
		return ret;
	return reg_write(icd, reg, ret | data);
}

static int reg_clear(struct soc_camera_device *icd, const u8 reg,
		     const u8 data)
{
	int ret;

	ret = reg_read(icd, reg);
	if (ret < 0)
		return ret;
	return reg_write(icd, reg, ret & ~data);
}

static int ov7620_reset(struct soc_camera_device *icd)
{
	int ret;

	ret = reg_set(icd, OV7620_COMA, OV7620_COMA7);
	if (ret < 0)
		return ret;
	
	mdelay(10);
	ret = reg_read(icd, OV7620_COMA);
	if (ret < 0)
		return ret;

	if (ret & OV7620_COMA7)
		dev_warn(&icd->dev, "Unable to reset OV7620 chip");

	return 0;
}

static int ov7620_power(struct soc_camera_device *icd, int up)
{
	if (up)
		return reg_clear(icd, OV7620_COMO, OV7620_COMO5);
	else
		return reg_set(icd, OV7620_COMO, OV7620_COMO5);
}

static int ov7620_init(struct soc_camera_device *icd, unsigned int flags)
{
/* 	struct ov7620 *ov7620 = container_of(icd, struct ov7620, icd); */
	int ret;
	dev_dbg(&icd->dev, "ov7620_init begin\n");

	/* Soft reset */
	ret = ov7620_reset(icd);
	if (ret < 0)
		return ret;

	/* process flags */
	if (!(flags & IS_PCLK_SAMPLE_RISING))
		ret = reg_set(icd, OV7620_COMD, OV7620_COMD6);
	if (ret < 0)
		return ret;

	if (!(flags & IS_HSYNC_ACTIVE_HIGH))
		ret = reg_set(icd, OV7620_COMC, OV7620_COMC3);
	if (ret < 0)
		return ret;

	if (!(flags & IS_VSYNC_ACTIVE_HIGH))
		ret = reg_set(icd, OV7620_CLK, OV7620_SYN6);
	if (ret < 0)
		return ret;

	if (flags & IS_MASTER)
		/* pixel clock gated by HREF */
		ret = reg_set(icd, OV7620_COML, OV7620_COML6);
	else
		ret = reg_set(icd, OV7620_COMI, OV7620_COMI6);
	if (ret < 0)
		return ret;

	/* Data multiplexed on Y bus */
	ret = reg_set(icd, OV7620_COMB, OV7620_COMB5);

	/* Progressive scan */
	ret = reg_set(icd, OV7620_COMH, OV7620_COMH5);

	/* No output */
	ret = reg_read(icd, OV7620_FD);
	if (ret < 0)
		return ret;
	else
		ret = reg_write(icd, OV7620_FD, (ret & ~OV7620_FD10));
	if (ret < 0)
		return ret;
	
	/* turn sensor off again because reset turned it on */
	ret = ov7620_power(icd, 0);
	if (ret < 0)
		return ret;

	dev_dbg(&icd->dev, "ov7620_init end\n");
	return 0;
}

static int ov7620_release(struct soc_camera_device *icd)
{
	/* nothing to do: sensor turned off with stream off */
	return 0;
}

static int ov7620_start_capture(struct soc_camera_device *icd)
{
	int ret;

	/* turn sensor on */
	ret = ov7620_power(icd, 1);
	if (ret < 0)
		return ret;

	/* enable frame drop */
	ret = reg_read(icd, OV7620_FD);
	if (ret < 0)
		return ret;
	else
		ret = reg_write(icd, OV7620_FD, (ret | OV7620_FD10));
	if (ret < 0)
		return ret;

	return 0;
}

static int ov7620_stop_capture(struct soc_camera_device *icd)
{
	int ret;

	/* No output */
	ret = reg_read(icd, OV7620_FD);
	if (ret < 0)
		return ret;
	else
		ret = reg_write(icd, OV7620_FD, (ret & ~OV7620_FD10));
	if (ret < 0)
		return ret;
	
	/* turn sensor off */
	ret = ov7620_power(icd, 0);
	if (ret < 0)
		return ret;

	return 0;
}

static int ov7620_set_capture_format(struct soc_camera_device *icd,
		__u32 pixfmt, struct v4l2_rect *rect, unsigned int flags)
{
	unsigned int width_flag = flags & (IS_DATAWIDTH_10 | IS_DATAWIDTH_9 |
					   IS_DATAWIDTH_8);
	int ret = 0;
	int shift_h, shift_v;
	u8 h_start, v_start;
	dev_dbg(&icd->dev, "ov7620_set_capture_format begin\n");

	/* Only one width bit may be set */
	if (!is_power_of_2(width_flag))
		return -EINVAL;

	/* Pixel format */
	switch (pixfmt) {
	case V4L2_PIX_FMT_YUV422P:
	case V4L2_PIX_FMT_YUYV:
	case V4L2_PIX_FMT_UYVY:
		/* YUV output format */ 
		ret = reg_clear(icd, OV7620_COMA, OV7620_COMA3);
		if (ret >= 0)
			ret = reg_set(icd, OV7620_COMB, OV7620_COMB5);
		if (ret >= 0)
			ret = reg_set(icd, OV7620_EHSH, OV7620_EHSH4);
		if (ret >= 0)
			ret = reg_clear(icd, OV7620_YUV, OV7620_YUV3);
		if (ret >= 0)
			ret = reg_set(icd, OV7620_YUV, OV7620_YUV2);
		if (ret >= 0)
			ret = reg_set(icd, OV7620_SPCB, OV7620_SPCB7);
		if (ret >= 0)
			ret = reg_clear(icd, OV7620_COMG, OV7620_COMG4);
		if (ret >= 0) {
			if (pixfmt == V4L2_PIX_FMT_YUYV)
				ret = reg_set(icd, OV7620_COMA, OV7620_COMA4);
			else
				ret = reg_clear(icd, OV7620_COMA, OV7620_COMA4);
		}
		break;
	case V4L2_PIX_FMT_GREY:
		/* YUV output format */ 
		ret = reg_clear(icd, OV7620_COMA, OV7620_COMA3);
		if (ret >= 0)
			/* particular Y8 case: only Y data output 
			   no Y/UV multiplexing on Y bus */
			ret = reg_clear(icd, OV7620_COMB, OV7620_COMB5);
		if (ret >= 0)
			ret = reg_clear(icd, OV7620_YUV, OV7620_YUV3);
		if (ret >= 0)
			ret = reg_clear(icd, OV7620_YUV, OV7620_YUV2);
		if (ret >= 0)
			ret = reg_set(icd, OV7620_SPCB, OV7620_SPCB7);
		if (ret >= 0)
			ret = reg_clear(icd, OV7620_COMG, OV7620_COMG4);
		break;
	case V4L2_PIX_FMT_GRB422P:
	case V4L2_PIX_FMT_SBGGR8:
		ret = reg_set(icd, OV7620_COMA, OV7620_COMA3);
		if (ret >= 0)
			ret = reg_set(icd, OV7620_COMB, OV7620_COMB5);
		if (ret >= 0)
			ret = reg_clear(icd, OV7620_EHSH, OV7620_EHSH4);
		if (ret >= 0)
			ret = reg_clear(icd, OV7620_SPCB, OV7620_SPCB7);
		if (ret >= 0)
			ret = reg_clear(icd, OV7620_COMH, 0x80);
		if (ret >= 0)
			ret = reg_clear(icd, OV7620_COMH, 0x04);
		if (ret >= 0)
			ret = reg_set(icd, OV7620_COMG, OV7620_COMG4);
		if (ret >= 0)
			ret = reg_write(icd, OV7620_YUV, 0x00);
		break;
	case 0:
		/* No format change, only geometry */
		break;
	default:
		return -EINVAL;
	}

	if (ret < 0) return ret;

	/* Setup frame format */
	if (icd->width_max == OV7620_W_MAX) {
		/* VGA output format */
		shift_h = 2;
		shift_v = 1;
	}
	else {
		/* QVGA output format */
		shift_h = 1;
		shift_v = 0;
	}
	h_start = OV7620_HW_MIN + (rect->left >> shift_h);
	ret = reg_write(icd, OV7620_HSTART, h_start);
	if (ret >= 0)
		ret = reg_write(icd, OV7620_HSTOP,
				h_start + (rect->width >> shift_h));
	v_start = OV7620_VW_MIN + (rect->top >> shift_v);
	if (ret >= 0)
		ret = reg_write(icd, OV7620_VSTART, v_start);
	if (ret >= 0)
		ret = reg_write(icd, OV7620_VSTOP,
				v_start + (rect->height >> shift_v));
	if (ret < 0)
		return ret;

	dev_dbg(&icd->dev, "ov7620_set_capture_format end\n");
	return 0;
}

static int ov7620_try_fmt_cap(struct soc_camera_device *icd,
			       struct v4l2_format *f)
{
	int shift;

	if (icd->width_max == OV7620_W_MAX)
		/* VGA output format */
		shift = 0;
	else
		/* QVGA output format */
		shift = 1;

	if (f->fmt.pix.height < (OV7620_H_MIN >> shift))
		f->fmt.pix.height = OV7620_W_MIN >> shift;
	if (f->fmt.pix.height > (OV7620_H_MAX >> shift))
		f->fmt.pix.height = OV7620_H_MAX >> shift;
	if (f->fmt.pix.width < (OV7620_W_MIN >> shift))
		f->fmt.pix.width = OV7620_W_MIN >> shift;
	if (f->fmt.pix.width > (OV7620_W_MAX >> shift))
		f->fmt.pix.width = OV7620_W_MAX >> shift;

	return 0;
}

static int ov7620_get_chip_id(struct soc_camera_device *icd,
			       struct v4l2_chip_ident *id)
{
	struct ov7620 *ov7620 = container_of(icd, struct ov7620, icd);

	if (id->match_type != V4L2_CHIP_MATCH_I2C_ADDR)
		return -EINVAL;

	if (id->match_chip != ov7620->client->addr)
		return -ENODEV;

	id->ident	= ov7620->model;
	id->revision	= 0;

	return 0;
}

#ifdef CONFIG_VIDEO_ADV_DEBUG
static int ov7620_get_register(struct soc_camera_device *icd,
				struct v4l2_register *reg)
{
	struct ov7620 *ov7620 = container_of(icd, struct ov7620, icd);

	if (reg->match_type != V4L2_CHIP_MATCH_I2C_ADDR || reg->reg > 0xff)
		return -EINVAL;

	if (reg->match_chip != ov7620->client->addr)
		return -ENODEV;

	reg->val = reg_read(icd, reg->reg);

	if (reg->val > 0xff)
		return -EIO;

	return 0;
}

static int ov7620_set_register(struct soc_camera_device *icd,
				struct v4l2_register *reg)
{
	struct ov7620 *ov7620 = container_of(icd, struct ov7620, icd);

	if (reg->match_type != V4L2_CHIP_MATCH_I2C_ADDR || reg->reg > 0xff)
		return -EINVAL;

	if (reg->match_chip != ov7620->client->addr)
		return -ENODEV;

	if (reg_write(icd, reg->reg, reg->val) < 0)
		return -EIO;

	return 0;
}
#endif

static unsigned int ov7620_get_datawidth(struct soc_camera_device *icd)
{
	struct ov7620 *ov7620 = container_of(icd, struct ov7620, icd);
	return ov7620->datawidth;
}

const struct v4l2_queryctrl ov7620_controls[] = {
	{
		.id		= V4L2_CID_BRIGHTNESS,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Brightness control (reg 0x06)",
		.minimum	= 0,
		.maximum	= 255,
		.step		= 1,
		.default_value	= 128,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_SATURATION,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Saturation control (reg 0x03)",
		.minimum	= 0,
		.maximum	= 255,
		.step		= 1,
		.default_value	= 128,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_AUTO_WHITE_BALANCE,
		.type		= V4L2_CTRL_TYPE_BOOLEAN,
		.name		= "AWB Auto White Balance (COMA2)",
		.minimum	= 0,
		.maximum	= 1,
		.step		= 1,
		.default_value	= 1,
	}, {
		.id		= V4L2_CID_DO_WHITE_BALANCE,
		.type		= V4L2_CTRL_TYPE_BUTTON,
		.name		= "Do white balance (COMA2)",
		.minimum	= 0,
		.maximum	= 0,
		.step		= 0,
		.default_value	= 0,
	}, {
		.id		= V4L2_CID_RED_BALANCE,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Red gain control (reg 0x02)",
		.minimum	= 0,
		.maximum	= 255,
		.step		= 1,
		.default_value	= 128,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_BLUE_BALANCE,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Blue gain control (reg 0x02)",
		.minimum	= 0,
		.maximum	= 255,
		.step		= 1,
		.default_value	= 128,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_GAMMA,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Gamma control (reg 0x64)",
		.minimum	= 0,
		.maximum	= 255,
		.step		= 1,
		.default_value	= 0x59,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_EXPOSURE,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Exposure control (reg 0x10)",
		.minimum	= 0,
		.maximum	= 255,
		.step		= 1,
		.default_value	= 255,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_AUTOGAIN,
		.type		= V4L2_CTRL_TYPE_BOOLEAN,
		.name		= "Automatic Gain (COMA5)",
		.minimum	= 0,
		.maximum	= 1,
		.step		= 1,
		.default_value	= 1,
	}, {
		.id		= V4L2_CID_GAIN,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "AGC gain control",
		.minimum	= 0,
		.maximum	= 63,
		.step		= 1,
		.default_value	= 0,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_SHARPNESS,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Sharpness control (SHP<3:0>)",
		.minimum	= 0,
		.maximum	= 15,
		.step		= 1,
		.default_value	= 3,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_EXPOSURE_AUTO,
		.type		= V4L2_CTRL_TYPE_BOOLEAN,
		.name		= "AEC Auto Exposure (COMI5)",
		.minimum	= 0,
		.maximum	= 1,
		.step		= 1,
		.default_value	= 1,
	}, {
		.id		= OV7620_CID_AUTOMODE,
		.type		= V4L2_CTRL_TYPE_BOOLEAN,
		.name		= "Auto adjust mode (COMB0)",
		.minimum	= 0,
		.maximum	= 1,
		.step		= 1,
		.default_value	= 1,
	}, {
		.id		= OV7620_CID_QVGA,
		.type		= V4L2_CTRL_TYPE_BOOLEAN,
		.name		= "QVGA resolution (COMC5)",
		.minimum	= 0,
		.maximum	= 1,
		.step		= 1,
		.default_value	= 1,
	}
};

static int ov7620_get_control(struct soc_camera_device *icd,
			       struct v4l2_control *ctrl);
static int ov7620_set_control(struct soc_camera_device *icd,
			       struct v4l2_control *ctrl);

static struct soc_camera_ops ov7620_ops = {
	.owner			= THIS_MODULE,
	.init			= ov7620_init,
	.release		= ov7620_release,
	.start_capture		= ov7620_start_capture,
	.stop_capture		= ov7620_stop_capture,
	.set_capture_format	= ov7620_set_capture_format,
	.try_fmt_cap		= ov7620_try_fmt_cap,
	.formats		= ov7620_data_formats,
	.num_formats		= ARRAY_SIZE(ov7620_data_formats),
	.get_datawidth		= ov7620_get_datawidth,
	.controls		= ov7620_controls,
	.num_controls		= ARRAY_SIZE(ov7620_controls),
	.get_control		= ov7620_get_control,
	.set_control		= ov7620_set_control,
	.get_chip_id		= ov7620_get_chip_id,
#ifdef CONFIG_VIDEO_ADV_DEBUG
	.get_register		= ov7620_get_register,
	.set_register		= ov7620_set_register,
#endif
};

static int ov7620_get_control(struct soc_camera_device *icd,
			       struct v4l2_control *ctrl)
{
	int ret;
	int data;

	ret = 0;

	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		data = reg_read(icd, OV7620_BRT);
		if (data < 0)
			return -EIO;
		ctrl->value = data;
		break;
	case V4L2_CID_SATURATION:
		data = reg_read(icd, OV7620_SAT);
		if (data < 0)
			return -EIO;
		ctrl->value = data;
		break;
	case V4L2_CID_AUTO_WHITE_BALANCE:
		data = reg_read(icd, OV7620_COMA);
		if (data < 0)
			return -EIO;
		ctrl->value =  !!(data & OV7620_COMA2);
		break;
/* 	case V4L2_CID_DO_WHITE_BALANCE: */
/* 		/\* not implemented: is a button *\/ */
/* 		break; */
	case V4L2_CID_RED_BALANCE:
		data = reg_read(icd, OV7620_RED);
		if (data < 0)
			return -EIO;
		ctrl->value = data;
		break;
	case V4L2_CID_BLUE_BALANCE:
		data = reg_read(icd, OV7620_BLUE);
		if (data < 0)
			return -EIO;
		ctrl->value = data;
		break;
	case V4L2_CID_GAMMA:
		data = reg_read(icd, OV7620_YGM);
		if (data < 0)
			return -EIO;
		ctrl->value = data;
		break;
	case V4L2_CID_EXPOSURE:
		data = reg_read(icd, OV7620_AEC);
		if (data < 0)
			return -EIO;
		ctrl->value = data;
		break;
	case V4L2_CID_AUTOGAIN:
		data = reg_read(icd, OV7620_COMA);
		if (data < 0)
			return -EIO;
		ctrl->value = !!(data & OV7620_COMA5);
		break;
	case V4L2_CID_GAIN:
		data = reg_read(icd, OV7620_AGC);
		if (data < 0)
			return -EIO;
		ctrl->value = data;
		break;
	case V4L2_CID_SHARPNESS:
		data = reg_read(icd, OV7620_SHP);
		if (data < 0)
			return -EIO;
		/* this controls magnitude only */
		ctrl->value = data & OV7620_SHP_MSK;
		break;
	case V4L2_CID_EXPOSURE_AUTO:
		data = reg_read(icd, OV7620_COMI);
		if (data < 0)
			return -EIO;
		/* !! negative logic */
		ctrl->value = !(data & OV7620_COMI7);
		break;
	case OV7620_CID_AUTOMODE:
		data = reg_read(icd, OV7620_COMB);
		if (data < 0)
			return -EIO;
		ctrl->value = !!(data & OV7620_COMB0);
		break;
	case OV7620_CID_QVGA:
		data = reg_read(icd, OV7620_COMC);
		if (data < 0)
			return -EIO;
		ctrl->value = !!(data & OV7620_COMC5);
		break;	default :
		ret = -EINVAL;
	}

	return ret;
}

static int ov7620_set_control(struct soc_camera_device *icd,
			       struct v4l2_control *ctrl)
{
	int ret;
	int data;
	const struct v4l2_queryctrl *qctrl;

	qctrl = soc_camera_find_qctrl(&ov7620_ops, ctrl->id);

	if (!qctrl)
		return -EINVAL;

	if (ctrl->value > qctrl->maximum || ctrl->value < qctrl->minimum)
		return -EINVAL;

	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		ret = reg_write(icd, OV7620_BRT, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_SATURATION:
		ret = reg_write(icd, OV7620_SAT, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_AUTO_WHITE_BALANCE:
		if (ctrl->value)
			ret = reg_set(icd, OV7620_COMA, OV7620_COMA2);
		else
			ret = reg_clear(icd, OV7620_COMA, OV7620_COMA2);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_DO_WHITE_BALANCE:
 		/* turn AWB on then off after a while */
		ret = reg_set(icd, OV7620_COMA, OV7620_COMA2);
		if (ret < 0)
			return -EIO;
		msleep(200);
		ret = reg_clear(icd, OV7620_COMA, OV7620_COMA2);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_RED_BALANCE:
		ret = reg_write(icd, OV7620_RED, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_BLUE_BALANCE:
		ret = reg_write(icd, OV7620_BLUE, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_GAMMA:
		ret = reg_write(icd, OV7620_YGM, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_EXPOSURE:
		ret = reg_write(icd, OV7620_AEC, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_AUTOGAIN:
		if (ctrl->value)
			ret = reg_set(icd, OV7620_COMA, OV7620_COMA5);
		else
			ret = reg_clear(icd, OV7620_COMA, OV7620_COMA5);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_GAIN:
		ret = reg_write(icd, OV7620_AGC, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_SHARPNESS:
		data = reg_read(icd, OV7620_SHP);
		if (data < 0)
			return -EIO;
		/* this controls magnitude only */
		data = (data & ~OV7620_SHP_MSK) | ctrl->value;
		ret = reg_write(icd, OV7620_AGC, (u8) data);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_EXPOSURE_AUTO:
		/* !! negative logic */
		if (ctrl->value)
			ret = reg_clear(icd, OV7620_COMI, OV7620_COMI7);
		else
			ret = reg_set(icd, OV7620_COMI, OV7620_COMI7);
		if (ret < 0)
			return -EIO;
		break;
	case OV7620_CID_AUTOMODE:
		if (ctrl->value)
			ret = reg_set(icd, OV7620_COMB, OV7620_COMB0);
		else
			ret = reg_clear(icd, OV7620_COMB, OV7620_COMB0);
		if (ret < 0)
			return -EIO;
		break;
	case OV7620_CID_QVGA:
		if (ctrl->value) {
			/* switch to QVGA output format */
			ret = reg_set(icd, OV7620_COMC, OV7620_COMC5);
			/* update capabilities */
			icd->width_def  = OV7620_W_DEF >> 1;
			icd->height_def = OV7620_H_DEF >> 1;
			icd->x_def      = (OV7620_HWS_DEF-OV7620_HW_MIN)*2;
			icd->y_def      = (OV7620_VWS_DEF-OV7620_VW_MIN)*1;
			icd->width      = icd->width_def;
			icd->height     = icd->height_def;
			icd->x_current	= icd->x_def;
			icd->y_current	= icd->y_def;
			icd->width_min  = OV7620_W_MIN >> 1;
			icd->width_max  = OV7620_W_MAX >> 1;
			icd->height_min	= OV7620_H_MIN >> 1;
			icd->height_max	= OV7620_H_MAX >> 1;
			icd->x_min	= 0;
			icd->y_min	= 0;
		}
		else {
			/* switch to VGA output format */
			ret = reg_clear(icd, OV7620_COMC, OV7620_COMC5);
			/* update capabilities */
			icd->width_def  = OV7620_W_DEF;
			icd->height_def = OV7620_H_DEF;
			icd->x_def      = (OV7620_HWS_DEF-OV7620_HW_MIN)*4;
			icd->y_def      = (OV7620_VWS_DEF-OV7620_VW_MIN)*2;
			icd->width      = icd->width_def;
			icd->height     = icd->height_def;
			icd->x_current	= icd->x_def;
			icd->y_current	= icd->y_def;
			icd->width_min  = OV7620_W_MIN;
			icd->width_max  = OV7620_W_MAX;
			icd->height_min	= OV7620_H_MIN;
			icd->height_max	= OV7620_H_MAX;
			icd->x_min	= 0;
			icd->y_min	= 0;
		}
		/* TODO set_capture_format */
		
		if (ret < 0)
			return -EIO;
		break;
	default :
		ret = -EINVAL;
	}
	return 0;
}

static int ov7620_video_probe(struct soc_camera_device *icd)
{
	int ret;
	dev_dbg(&icd->dev, "ov7620_video_probe begin\n");

	if (!icd->dev.parent ||
	    to_soc_camera_host(icd->dev.parent)->nr != icd->iface)
		return -ENODEV;

	/* everything is OK, sensor was probed in ov7620_detect_client */

	ret = soc_camera_video_start(icd);
	if (ret < 0)
		return ret;

	dev_info(&icd->dev, "Detected an OV7620 chip\n");

	dev_dbg(&icd->dev, "ov7620_video_probe end\n");
	return 0;
}

static void ov7620_video_remove(struct soc_camera_device *icd)
{
	struct ov7620 *ov7620 = container_of(icd, struct ov7620, icd);

	/* turn sensor off */
	ov7620_power(icd, 0);

	dev_dbg(&icd->dev, "Video %x removed: %p, %p\n", ov7620->client->addr,
		icd->dev.parent, icd->vdev);
	soc_camera_video_stop(icd);
}


/*
 *****************************************************************************
 *
 *	platform_data Driver Interface    (get I2C address from platform data)
 *
 *****************************************************************************
 */
static int __init ov7620_platform_probe(struct platform_device *pdev)
{
	struct ov7620_platform_data *pdata;

	pdata = pdev->dev.platform_data;
	if (!pdev || !pdata) {
	  return -ENODEV;
	}

	normal_addr[0] = pdata->i2c_addr;
	ov7620_ctx.icl = &pdata->iclink;

	return 0;
}

static int ov7620_platform_remove(struct platform_device *pdev)
{
	return 0;
}

static struct platform_driver ov7620_platform_driver = {
	.probe  = ov7620_platform_probe,
	.remove = ov7620_platform_remove,
	.driver = {
		.owner = THIS_MODULE,
		.name  = OV7620_DRV_NAME,
	},
};

/*
 *****************************************************************************
 *
 *	I2C Driver Interface
 *
 *****************************************************************************
 */
static int ov7620_detect_client (struct i2c_adapter *adap, int addr, int kind)
{
	s32 data_h, data_l;
	struct i2c_client *client;
	struct soc_camera_device *icd;
	int rc;

	if (!i2c_check_functionality(adap, I2C_FUNC_I2C
			| I2C_FUNC_SMBUS_BYTE_DATA))
		return -EIO;

	client = kzalloc(sizeof(struct i2c_client), GFP_KERNEL);
	if (!client)
		return -ENOMEM;

	strlcpy(client->name, "OV7620", I2C_NAME_SIZE);
	client->addr = addr;
	client->adapter = adap;
	client->driver = &ov7620_i2c_driver;

	if ((rc = i2c_attach_client(client)))
		goto attach_err;

	ov7620_ctx.client = client;
	i2c_set_clientdata(client, &ov7620_ctx);

	icd = &ov7620_ctx.icd;

	/* Read out the chip manufacture ID registers */
	data_h = reg_read(icd, OV7620_MIDH);
	data_l = reg_read(icd, OV7620_MIDL);
	if (data_h != OV7620_MIDH_VAL || data_l != OV7620_MIDL_VAL) {
		rc = -ENODEV;
		dev_info(&icd->dev, "No OV7620 detected, MIDH=0x%x, MIDL=0x%x\n",
			 data_h, data_l);
		goto detach_client;
	}
	
	/* turn sensor off until activation */
	rc = ov7620_power(icd, 0);
	if (rc < 0)
		goto detach_client;

	icd->probe	= ov7620_video_probe;
	icd->remove	= ov7620_video_remove;
	icd->ops	= &ov7620_ops;
	icd->control	= &client->dev;
	icd->width_def  = OV7620_W_DEF;
	icd->height_def = OV7620_H_DEF;
	icd->x_def      = (OV7620_HWS_DEF-OV7620_HW_MIN)*4;
	icd->y_def      = (OV7620_VWS_DEF-OV7620_VW_MIN)*2;
	icd->width      = icd->width_def;
	icd->height     = icd->height_def;
	icd->x_current	= icd->x_def;
	icd->y_current	= icd->y_def;
	icd->x_min	= 0;
	icd->y_min	= 0;
	icd->width_min	= OV7620_W_MIN;
	icd->width_max	= OV7620_W_MAX;
	icd->height_min	= OV7620_H_MIN;
	icd->height_max	= OV7620_H_MAX;
	icd->y_skip_top	= 0;
	icd->iface	= ov7620_ctx.icl->bus_id;
	ov7620_ctx.datawidth = 8;
	ov7620_ctx.model = V4L2_IDENT_OV7620;

	rc = soc_camera_device_register(icd);
	if (rc)
		goto detach_client;

	return 0;

detach_client:
	i2c_detach_client(client);
attach_err:
	kfree(client);
	return rc;
}

static int ov7620_attach(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data, ov7620_detect_client);
}

static int ov7620_detach(struct i2c_client *client)
{
	struct ov7620 *ov7620 = i2c_get_clientdata(client);
	int rc;

	dev_info(&client->dev, "ov7620_detach\n");

	soc_camera_device_unregister(&ov7620->icd);

	rc = i2c_detach_client(client);
	if (rc == 0) {
		kfree(client);
	}

	return rc;
}

static struct i2c_driver ov7620_i2c_driver = {
	.driver = {
		.name = "ov7620",
	},
	.id 		= I2C_DRIVERID_OV7620,
/* 	.class 		= I2C_CLASS_CAM_DIGITAL, */
	.attach_adapter = ov7620_attach,
	.detach_client	= ov7620_detach,
/* 	.probe		= ov7620_probe, */
/* 	.remove		= ov7620_remove, */
};

static int __init ov7620_mod_init(void)
{
	int ret;
	ret = platform_driver_register(&ov7620_platform_driver);
	if (ret == 0)
		ret = i2c_add_driver(&ov7620_i2c_driver);

	return ret;
}

static void __exit ov7620_mod_exit(void)
{
	i2c_del_driver(&ov7620_i2c_driver);
	platform_driver_unregister(&ov7620_platform_driver);
}

module_init(ov7620_mod_init);
module_exit(ov7620_mod_exit);

MODULE_DESCRIPTION("Omnivision OV7620 Camera driver with soc_camera bus");
MODULE_AUTHOR("Florian Pantaleao <fpantaleao@mobisensesystems.com>");
MODULE_LICENSE("GPL");
