/*
 * Omnivision OV7720 Driver with soc_camera bus
 *
 * Copyright (C) 2008, Florian Pantaleao <fpantaleao@mobisensesystems.com>
 *
 * Inspired from Guennadi Liakhovetski mt9v022.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* #define DEBUG 1 */
#include <linux/device.h>

#include <linux/videodev2.h>
#include <linux/ov7720.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/log2.h>
#include <linux/platform_device.h>

#include <media/v4l2-common.h>
#include <media/v4l2-chip-ident.h>
#include <media/soc_camera.h>
#include <media/soc_ov7720.h>

/* OV7720 i2c address 0x21 */

/* static char *sensor_type; */
/* module_param(sensor_type, charp, S_IRUGO); */
/* MODULE_PARM_DESC(sensor_type, "Sensor type: \"colour\" or \"monochrome\"\n"); */

static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short normal_addr[] = { OV7720_I2C_ADDR, I2C_CLIENT_END };

static struct i2c_client_address_data addr_data = {
	.normal_i2c	= normal_addr,
	.probe		= ignore,
	.ignore		= ignore,
};

/* Forward declarations */
static struct i2c_driver ov7720_i2c_driver;

/* Manufacture ID constants */
#define OV7720_MIDH_VAL		0x7F
#define OV7720_MIDL_VAL		0xA2


/* OV7720 selected register addresses */
#define OV7720_AGC		0x0000
#define OV7720_BLUE		0x0001
#define OV7720_RED		0x0002
#define OV7720_GREEN		0x0003
#define OV7720_BAVG		0x0005
#define OV7720_GAVG		0x0006
#define OV7720_RAVG		0x0007
#define OV7720_AECH		0x0008
#define OV7720_COM2             0x0009
#define OV7720_PID              0x000A
#define OV7720_VER              0x000B
#define OV7720_COM3		0x000C
#define OV7720_COM4		0x000D
#define OV7720_COM5		0x000E
#define OV7720_COM6		0x000F
#define OV7720_AEC		0x0010
#define OV7720_CLKRC		0x0011
#define OV7720_COM7		0x0012
#define OV7720_COM8		0x0013
#define OV7720_COM9		0x0014
#define OV7720_COM10		0x0015
#define OV7720_REG16		0x0016
#define OV7720_HSTART		0x0017
#define OV7720_HSIZE		0x0018
#define OV7720_VSTART		0x0019
#define OV7720_VSIZE		0x001A
#define OV7720_PSHFT		0x001B
#define OV7720_MIDH		0x001C		// Manufacture ID High byte
#define OV7720_MIDL		0x001D		// Manufacture ID Low byte
#define OV7720_LAEC		0x001E
#define OV7720_COM11		0x0020
#define OV7720_BDBASE		0x0022
#define OV7720_BDMSTEP		0x0023
#define OV7720_AEW		0x0024
#define OV7720_AEB		0x0025
#define OV7720_VPT		0x0026
#define OV7720_REG28		0x0028
#define OV7720_HOUTSIZE		0x0029
#define OV7720_EXHCH		0x002A
#define OV7720_EXHCL		0x002B
#define OV7720_VOUTSIZE		0x002C
#define OV7720_ADVFL		0x002D
#define OV7720_ADVFH		0x002E
#define OV7720_LUMHTH		0x0030
#define OV7720_LUMLTH		0x0031
#define OV7720_HREF		0x0032
#define OV7720_DM_LNL		0x0033
#define OV7720_DM_LNH		0x0034
#define OV7720_ADOFF_B		0x0035
#define OV7720_ADOFF_R		0x0036
#define OV7720_ADOFF_GB		0x0037
#define OV7720_ADOFF_GR		0x0038
#define OV7720_OFF_B		0x0039
#define OV7720_OFF_R		0x003A
#define OV7720_OFF_GB		0x003B
#define OV7720_OFF_GR		0x003C
#define OV7720_COM12		0x003D
#define OV7720_COM13		0x003E
#define OV7720_COM14		0x003F
#define OV7720_COM15		0x0040
#define OV7720_COM16		0x0041
#define OV7720_TGT_B		0x0042
#define OV7720_TGT_R		0x0043
#define OV7720_TGT_GB		0x0044
#define OV7720_TGT_GR		0x0045
#define OV7720_LC_CTR		0x0046
#define OV7720_LC_XC		0x0047
#define OV7720_LC_YC		0x0048
#define OV7720_LC_COEF		0x0049
#define OV7720_LC_RADI		0x004A
#define OV7720_LC_COEFB		0x004B
#define OV7720_LC_COEFR		0x004C
#define OV7720_FIXGAIN		0x004D
#define OV7720_AREF0		0x004E
#define OV7720_AREF1		0x004F
#define OV7720_AREF2		0x0050
#define OV7720_AREF3		0x0051
#define OV7720_AREF4		0x0052
#define OV7720_AREF5		0x0053
#define OV7720_AREF6		0x0054
#define OV7720_AREF7		0x0055
#define OV7720_UFIX		0x0060
#define OV7720_VFIX		0x0061
#define OV7720_AWBB_BLK		0x0062
#define OV7720_AWB_CTRL0	0x0063
#define OV7720_DSP_CTRL1	0x0064
#define OV7720_DSP_CTRL2	0x0065
#define OV7720_DSP_CTRL3	0x0066
#define OV7720_DSP_CTRL4	0x0067
#define OV7720_AWB_BIAS		0x0068
#define OV7720_AWBCTRL1		0x0069
#define OV7720_AWBCTRL2		0x006A
#define OV7720_AWBCTRL3		0x006B
#define OV7720_AWBCTRL4		0x006C
#define OV7720_AWBCTRL5		0x006D
#define OV7720_AWBCTRL6		0x006E
#define OV7720_AWBCTRL7		0x006F
#define OV7720_AWBCTRL8		0x0070
#define OV7720_AWBCTRL9		0x0071
#define OV7720_AWBCTRL10	0x0072
#define OV7720_AWBCTRL11	0x0073
#define OV7720_AWBCTRL12	0x0074
#define OV7720_AWBCTRL13	0x0075
#define OV7720_AWBCTRL14	0x0076
#define OV7720_AWBCTRL15	0x0077
#define OV7720_AWBCTRL16	0x0078
#define OV7720_AWBCTRL17	0x0079
#define OV7720_AWBCTRL18	0x007A
#define OV7720_AWBCTRL19	0x007B
#define OV7720_AWBCTRL20	0x007C
#define OV7720_AWBCTRL21	0x007D
#define OV7720_AWBCTRL21	0x007D
#define OV7720_GAM1		0x007E
#define OV7720_GAM2		0x007F
#define OV7720_GAM3		0x0080
#define OV7720_GAM4		0x0081
#define OV7720_GAM5		0x0082
#define OV7720_GAM6		0x0083
#define OV7720_GAM7		0x0084
#define OV7720_GAM8		0x0085
#define OV7720_GAM9		0x0086
#define OV7720_GAM10		0x0087
#define OV7720_GAM11		0x0088
#define OV7720_GAM12		0x0089
#define OV7720_GAM13		0x008A
#define OV7720_GAM14		0x008B
#define OV7720_GAM15		0x008C
#define OV7720_SLOP		0x008D
#define OV7720_DNSTH		0x008E
#define OV7720_EDGE0		0x008F
#define OV7720_EDGE1		0x0090
#define OV7720_DNSOFF		0x0091
#define OV7720_EDGE2		0x0092
#define OV7720_EDGE3		0x0093
#define OV7720_MTX1		0x0094
#define OV7720_MTX2		0x0095
#define OV7720_MTX3		0x0096
#define OV7720_MTX4		0x0097
#define OV7720_MTX5		0x0098
#define OV7720_MTX6		0x0099
#define OV7720_MTX_CTRL		0x009A
#define OV7720_BRIGHT		0x009B
#define OV7720_CNST		0x009C
#define OV7720_UVADJ0		0x009E
#define OV7720_UVADJ1		0x009F
#define OV7720_SCAL0		0x00A0
#define OV7720_SCAL1		0x00A1
#define OV7720_SCAL2		0x00A2
#define OV7720_FIFODLYM		0x00A3
#define OV7720_FIFODLYA		0x00A4
#define OV7720_SDE		0x00A6
#define OV7720_USAT		0x00A7
#define OV7720_VSAT		0x00A8
#define OV7720_HUECOS		0x00A9
#define OV7720_HUESIN		0x00AA
#define OV7720_SIGN		0x00AB
#define OV7720_DSPAUTO		0x00AC

// End of OV7720 register
#define OV7720_REGEND		( OV7720_AVG + 1 )

/* register bit definitions */
#define OV7720_COM2_SLEEP	0x10
#define OV7720_COM3_SWAP_YUV	0x10
#define OV7720_COM3_SWAP_BR	0x20
#define OV7720_COM7_RESET	0x80
#define OV7720_COM7_QVGA	0x40
#define OV7720_COM7_RAW		0x10
#define OV7720_COM7_RGB_FMT	0x0C
#define OV7720_COM7_OUT_FMT	0x03
#define OV7720_COM7_OUT_YUV	0x00
#define OV7720_COM7_OUT_RGB	0x02
#define OV7720_COM7_OUT_PBYR	0x01
#define OV7720_COM7_OUT_BYR	0x03
#define OV7720_COM8_AGC_EN	0x04
#define OV7720_COM8_AWB_EN	0x02
#define OV7720_COM8_AEC_EN	0x01
#define OV7720_COM10_HREF_REV	0x08
#define OV7720_COM10_VSYNC_NEG	0x02
#define OV7720_COM10_PCLK_OUT	0x20
#define OV7720_AREF1_REG_OFF	0x08
#define OV7720_DSP1_GAMMA_ON    0x04
#define OV7720_EDGE0_STRENGTH   0x1F
#define OV7720_SIGN_BRIGHT	0x03

/* OV7720 window parameters */
#define OV7720_HSTART_DEF_VGA 	0x0098 /* 152 */
#define OV7720_HSTART_DEF_QVGA 	0x00FC /* 252 */
#define OV7720_HSIZE_DEF_VGA 	0x0280 /* 640 */
#define OV7720_HSIZE_DEF_QVGA 	0x0140 /* 320 */
#define OV7720_VSTRT_DEF_VGA 	0x000E
#define OV7720_VSTRT_DEF_QVGA 	0x0006
#define OV7720_VSIZE_DEF_VGA 	0x01E0 /* 480 */
#define OV7720_VSIZE_DEF_QVGA 	0x00F0 /* 240 */

#define OV7720_W_MIN   4
#define OV7720_W_MAX   OV7720_HSIZE_DEF_VGA
#define OV7720_W_DEF   OV7720_HSIZE_DEF_VGA
#define OV7720_H_MIN   2
#define OV7720_H_MAX   OV7720_VSIZE_DEF_VGA
#define OV7720_H_DEF   OV7720_VSIZE_DEF_VGA


static const struct soc_camera_data_format ov7720_data_formats[] = {
	{
		.name		= "YUV422 planarized",
		.depth		= 16,
		.fourcc		= V4L2_PIX_FMT_YUV422P,
		.colorspace	= V4L2_COLORSPACE_JPEG,
	}, {
		.name		= "YUV422 packed UYVY",
		.depth		= 16,
		.fourcc		= V4L2_PIX_FMT_UYVY,
		.colorspace	= V4L2_COLORSPACE_JPEG,
	}, {
		.name		= "YUV422 packed YUYV",
		.depth		= 16,
		.fourcc		= V4L2_PIX_FMT_YUYV,
		.colorspace	= V4L2_COLORSPACE_JPEG,
	}, {
		.name		= "Bayer planarized 8 bit",
		.depth		= 16,
		.fourcc		= V4L2_PIX_FMT_GRB422P,
		.colorspace	= V4L2_COLORSPACE_SRGB,
	}, {
		.name		= "Bayer packed 8 bit",
		.depth		= 16,
		.fourcc		= V4L2_PIX_FMT_SBGGR8,
		.colorspace	= V4L2_COLORSPACE_SRGB,
	}
};

struct ov7720 {
	struct i2c_client *client;
	struct soc_camera_device icd;
	struct soc_camera_link *icl;
	int model;	/* V4L2_IDENT_OV7720 code from v4l2-chip-ident.h */
	unsigned char datawidth;
};

struct ov7720 ov7720_ctx;

static int reg_read(struct soc_camera_device *icd, const u8 reg)
{
	dev_dbg(&icd->dev, "reg_read 0x%02x\n", reg);
	struct ov7720 *ov7720 = container_of(icd, struct ov7720, icd);
/* 	return i2c_smbus_read_byte_data(ov7720->client, reg); */

	/* !!! SMBUS is not supported by OV7720 !!! */
	struct i2c_client *client = ov7720->client;
	char buf[1];

	buf[0] = reg;
        if (i2c_master_send(client, buf, 1) != 1) {
                dev_err(&icd->dev, "i2c_master_send failed.\n");
                return -1;
        }
        if (i2c_master_recv(client, buf, 1) != 1) {
                dev_err(&icd->dev, "i2c_master_recv failed.");
                return -1;
        }

	dev_dbg(&icd->dev, "reg_read 0x%02x = 0x%02x\n", reg, buf[0]);

        return buf[0];
}

static int reg_write(struct soc_camera_device *icd, const u8 reg,
		     const u8 data)
{
	dev_dbg(&icd->dev, "reg_write 0x%02x = 0x%02x\n", reg, data);
	struct ov7720 *ov7720 = container_of(icd, struct ov7720, icd);
	return i2c_smbus_write_byte_data(ov7720->client, reg, data);
}


static int reg_set(struct soc_camera_device *icd, const u8 reg,
		   const u8 data)
{
	int ret;

	ret = reg_read(icd, reg);
	if (ret < 0)
		return ret;
	return reg_write(icd, reg, ret | data);
}

static int reg_clear(struct soc_camera_device *icd, const u8 reg,
		     const u8 data)
{
	int ret;

	ret = reg_read(icd, reg);
	if (ret < 0)
		return ret;
	return reg_write(icd, reg, ret & ~data);
}

static int ov7720_reset(struct soc_camera_device *icd)
{
	int ret;

	ret = reg_set(icd, OV7720_COM7, OV7720_COM7_RESET);
	if (ret < 0)
		return ret;
	
	mdelay(10);
	ret = reg_read(icd, OV7720_COM7);
	if (ret < 0)
		return ret;

	if (ret != 0)
		dev_warn(&icd->dev, "Unable to reset OV7720 chip");

	return 0;
}

static int ov7720_power(struct soc_camera_device *icd, int up)
{
	int ret;

	if (up) {
		dev_dbg(&icd->dev, "ov7720_power ON\n");
		ret = reg_clear(icd, OV7720_COM2, OV7720_COM2_SLEEP);
		if (ret < 0)
			return ret;
		dev_dbg(&icd->dev, "Out of sleep\n");
	}
	else {
		dev_dbg(&icd->dev, "ov7720_power OFF\n");
		ret = reg_set(icd, OV7720_COM2, OV7720_COM2_SLEEP);
		if (ret < 0)
			return ret;
		dev_dbg(&icd->dev, "Soft sleep mode\n");
	}

	return 0;
}

static int ov7720_init(struct soc_camera_device *icd, unsigned int flags)
{
	int ret;
	dev_dbg(&icd->dev, "ov7720_init begin\n");

	/* Soft reset */
	ret = ov7720_reset(icd);
	if (ret < 0)
		return ret;

	ret = reg_read(icd, OV7720_COM7);
	ret = reg_read(icd, OV7720_HSTART);
	ret = reg_read(icd, OV7720_HSIZE);
	ret = reg_read(icd, OV7720_VSTART);
	ret = reg_read(icd, OV7720_VSIZE);

	/* process flags */
	if (!(flags & IS_HSYNC_ACTIVE_HIGH))
		ret = reg_set(icd, OV7720_COM10, OV7720_COM10_HREF_REV);
	if (ret < 0)
		return ret;

	if (!(flags & IS_VSYNC_ACTIVE_HIGH))
		ret = reg_set(icd, OV7720_COM10, OV7720_COM10_VSYNC_NEG);
	if (ret < 0)
		return ret;

	if (flags & IS_MASTER)
		/* pixel clock gated by HREF */
		ret = reg_set(icd, OV7720_COM10, OV7720_COM10_PCLK_OUT);
	else {
		ret = -1;
		dev_dbg(&icd->dev, "OV7720 can not run as slave\n");
	}
	if (ret < 0)
		return ret;
	
	/* CLKRC default setting: fastest, PCLK = 2*MCLK */
	/* COM2 default setting: lowest output drive capability */
	/* COM3 default setting: Y/UV swapped  */
	/* COM3: swap B/R for compatibility with OV7620 software */
	ret = reg_set(icd, OV7720_COM3, OV7720_COM3_SWAP_BR);
	if (ret < 0)
		return ret;

	/* turn sensor off */
	ret = ov7720_power(icd, 0);
	if (ret < 0)
		return ret;

	dev_dbg(&icd->dev, "ov7720_init end\n");
	return 0;
}

static inline int ov7720_release(struct soc_camera_device *icd)
{
	/* nothing to do with OV7720 */
	return 0;
}

static inline int ov7720_start_capture(struct soc_camera_device *icd)
{
	int ret;
	ret = ov7720_power(icd, 1);
	if (ret < 0)
		return ret;
	/* nothing to do with OV7720 */
	return 0;
}

static inline int ov7720_stop_capture(struct soc_camera_device *icd)
{
	int ret;
	ret = ov7720_power(icd, 0);
	if (ret < 0)
		return ret;
	/* nothing to do with OV7720 */
	return 0;
}

static int ov7720_set_output_format(struct soc_camera_device *icd,
				    int format, int yuv_swap)
{
	int ret;

	ret = reg_read(icd, OV7720_COM7);
	if (ret < 0)
		return ret;
	else {
		ret &= ~(OV7720_COM7_RAW |
			 OV7720_COM7_OUT_FMT |
			 OV7720_COM7_RGB_FMT);
		if ((format == OV7720_COM7_OUT_BYR) ||
		    (format == OV7720_COM7_OUT_PBYR))
			ret |= OV7720_COM7_RAW;
		ret |= format;
		ret = reg_write(icd, OV7720_COM7, ret);
		if (ret < 0)
			return ret;
	}
	if (yuv_swap)
		ret = reg_set(icd, OV7720_COM3, OV7720_COM3_SWAP_YUV);
	else
		ret = reg_clear(icd, OV7720_COM3, OV7720_COM3_SWAP_YUV);
	if (ret < 0)
		return ret;
	
	return 0;
}

static int ov7720_set_capture_format(struct soc_camera_device *icd,
		__u32 pixfmt, struct v4l2_rect *rect, unsigned int flags)
{
	unsigned int width_flag = flags & (IS_DATAWIDTH_10 | IS_DATAWIDTH_9 |
					   IS_DATAWIDTH_8);
	int ret = 0;
	int shift_h, shift_v;
	u8 h_start, v_start;
	dev_dbg(&icd->dev, "ov7720_set_capture_format begin\n");

	/* Only one width bit may be set */
	if (!is_power_of_2(width_flag))
		return -EINVAL;

	/* Pixel format */
	switch (pixfmt) {
	case V4L2_PIX_FMT_YUYV:
		/* YUV output format */ 
	  ret = ov7720_set_output_format(icd, OV7720_COM7_OUT_YUV, 0);
		break;
	case V4L2_PIX_FMT_UYVY:
	case V4L2_PIX_FMT_YUV422P:
		/* YUV output format */ 
	  ret = ov7720_set_output_format(icd, OV7720_COM7_OUT_YUV, 1);
		break;
	case V4L2_PIX_FMT_GRB422P:
		/* RGB output format */ 
	  ret = ov7720_set_output_format(icd, OV7720_COM7_OUT_RGB, 1);
		break;
	case V4L2_PIX_FMT_SBGGR8:
		/* Bayer output format */ 
/* 		ret = ov7720_set_output_format(icd, OV7720_COM7_OUT_PBYR, 0); */
	  ret = ov7720_set_output_format(icd, OV7720_COM7_OUT_RGB, 0);
		break;
	case 0:
		/* No format change, only geometry */
		break;
	default:
		return -EINVAL;
	}

	if (ret < 0) return ret;

/* 	/\* Setup frame format *\/ */
/* 	if (icd->width_max == OV7720_W_MAX) { */
/* 		/\* VGA output format *\/ */
/* 		shift_h = 2; */
/* 		shift_v = 1; */
/* 	} */
/* 	else { */
/* 		/\* QVGA output format *\/ */
/* 		shift_h = 1; */
/* 		shift_v = 0; */
/* 	} */
/* 	h_start = OV7720_HW_MIN + (rect->left >> shift_h); */
/* 	ret = reg_write(icd, OV7720_HSTART, h_start); */
/* 	if (ret >= 0) */
/* 		ret = reg_write(icd, OV7720_HSTOP, */
/* 				h_start + (rect->width >> shift_h)); */
/* 	v_start = OV7720_VW_MIN + (rect->top >> shift_v); */
/* 	if (ret >= 0) */
/* 		ret = reg_write(icd, OV7720_VSTART, v_start); */
/* 	if (ret >= 0) */
/* 		ret = reg_write(icd, OV7720_VSTOP, */
/* 				v_start + (rect->height >> shift_v)); */
/* 	if (ret < 0) */
/* 		return ret; */

	dev_dbg(&icd->dev, "ov7720_set_capture_format end\n");
	return 0;
}

static int ov7720_try_fmt_cap(struct soc_camera_device *icd,
			       struct v4l2_format *f)
{
	int shift;

	if (icd->width_max == OV7720_W_MAX)
		/* VGA output format */
		shift = 0;
	else
		/* QVGA output format */
		shift = 1;

	if (f->fmt.pix.height < (OV7720_H_MIN >> shift))
		f->fmt.pix.height = OV7720_W_MIN >> shift;
	if (f->fmt.pix.height > (OV7720_H_MAX >> shift))
		f->fmt.pix.height = OV7720_H_MAX >> shift;
	if (f->fmt.pix.width < (OV7720_W_MIN >> shift))
		f->fmt.pix.width = OV7720_W_MIN >> shift;
	if (f->fmt.pix.width > (OV7720_W_MAX >> shift))
		f->fmt.pix.width = OV7720_W_MAX >> shift;

	return 0;
}

static int ov7720_get_chip_id(struct soc_camera_device *icd,
			       struct v4l2_chip_ident *id)
{
	struct ov7720 *ov7720 = container_of(icd, struct ov7720, icd);

	if (id->match_type != V4L2_CHIP_MATCH_I2C_ADDR)
		return -EINVAL;

	if (id->match_chip != ov7720->client->addr)
		return -ENODEV;

	id->ident	= ov7720->model;
	id->revision	= 0;

	return 0;
}

#ifdef CONFIG_VIDEO_ADV_DEBUG
static int ov7720_get_register(struct soc_camera_device *icd,
				struct v4l2_register *reg)
{
	struct ov7720 *ov7720 = container_of(icd, struct ov7720, icd);

	if (reg->match_type != V4L2_CHIP_MATCH_I2C_ADDR || reg->reg > 0xff)
		return -EINVAL;

	if (reg->match_chip != ov7720->client->addr)
		return -ENODEV;

	reg->val = reg_read(icd, reg->reg);

	if (reg->val > 0xff)
		return -EIO;

	return 0;
}

static int ov7720_set_register(struct soc_camera_device *icd,
				struct v4l2_register *reg)
{
	struct ov7720 *ov7720 = container_of(icd, struct ov7720, icd);

	if (reg->match_type != V4L2_CHIP_MATCH_I2C_ADDR || reg->reg > 0xff)
		return -EINVAL;

	if (reg->match_chip != ov7720->client->addr)
		return -ENODEV;

	if (reg_write(icd, reg->reg, reg->val) < 0)
		return -EIO;

	return 0;
}
#endif

static unsigned int ov7720_get_datawidth(struct soc_camera_device *icd)
{
	struct ov7720 *ov7720 = container_of(icd, struct ov7720, icd);
	return ov7720->datawidth;
}

const struct v4l2_queryctrl ov7720_controls[] = {
	{
		.id		= V4L2_CID_BRIGHTNESS,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Brightness (0x9B and 0xAB[3])",
		.minimum	= 0,
		.maximum	= 512,
		.step		= 1,
		.default_value	= 256,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_SATURATION,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Saturation (0xA7 and 0xA8)",
		.minimum	= 0,
		.maximum	= 255,
		.step		= 1,
		.default_value	= 64,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_AUTO_WHITE_BALANCE,
		.type		= V4L2_CTRL_TYPE_BOOLEAN,
		.name		= "AWB Auto White Balance (0x13[1])",
		.minimum	= 0,
		.maximum	= 1,
		.step		= 1,
		.default_value	= 1,
	}, {
		.id		= V4L2_CID_DO_WHITE_BALANCE,
		.type		= V4L2_CTRL_TYPE_BUTTON,
		.name		= "Do white balance (0x13[1])",
		.minimum	= 0,
		.maximum	= 0,
		.step		= 0,
		.default_value	= 0,
	}, {
		.id		= V4L2_CID_RED_BALANCE,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Red gain control (reg 0x02)",
		.minimum	= 0,
		.maximum	= 255,
		.step		= 1,
		.default_value	= 128,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_BLUE_BALANCE,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Blue gain control (reg 0x02)",
		.minimum	= 0,
		.maximum	= 255,
		.step		= 1,
		.default_value	= 128,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_GAMMA,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Gamma enable (0x64[2])",
		.minimum	= 0,
		.maximum	= 1,
		.step		= 1,
		.default_value	= 1,
	}, {
		.id		= V4L2_CID_EXPOSURE,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Exposure control (0x08,0x10)",
		.minimum	= 0,
		.maximum	= 65535,
		.step		= 1,
		.default_value	= 64,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_AUTOGAIN,
		.type		= V4L2_CTRL_TYPE_BOOLEAN,
		.name		= "Automatic Gain (0x13[2])",
		.minimum	= 0,
		.maximum	= 1,
		.step		= 1,
		.default_value	= 1,
	}, {
		.id		= V4L2_CID_GAIN,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "AGC gain control (0x00)",
		.minimum	= 0,
		.maximum	= 255,
		.step		= 1,
		.default_value	= 0,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_SHARPNESS,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Sharpness control (0x8F[3:0])",
		.minimum	= 0,
		.maximum	= 31,
		.step		= 1,
		.default_value	= 0,
		.flags		= V4L2_CTRL_FLAG_SLIDER,
	}, {
		.id		= V4L2_CID_EXPOSURE_AUTO,
		.type		= V4L2_CTRL_TYPE_BOOLEAN,
		.name		= "AEC Auto Exposure (0x13[1])",
		.minimum	= 0,
		.maximum	= 1,
		.step		= 1,
		.default_value	= 1,
	}, {
		.id		= OV7720_CID_QVGA,
		.type		= V4L2_CTRL_TYPE_BOOLEAN,
		.name		= "QVGA resolution (0x12[6])",
		.minimum	= 0,
		.maximum	= 1,
		.step		= 1,
		.default_value	= 1,
	}
};

static int ov7720_get_control(struct soc_camera_device *icd,
			       struct v4l2_control *ctrl);
static int ov7720_set_control(struct soc_camera_device *icd,
			       struct v4l2_control *ctrl);

static struct soc_camera_ops ov7720_ops = {
	.owner			= THIS_MODULE,
	.init			= ov7720_init,
	.release		= ov7720_release,
	.start_capture		= ov7720_start_capture,
	.stop_capture		= ov7720_stop_capture,
	.set_capture_format	= ov7720_set_capture_format,
	.try_fmt_cap		= ov7720_try_fmt_cap,
	.formats		= ov7720_data_formats,
	.num_formats		= ARRAY_SIZE(ov7720_data_formats),
	.get_datawidth		= ov7720_get_datawidth,
	.controls		= ov7720_controls,
	.num_controls		= ARRAY_SIZE(ov7720_controls),
	.get_control		= ov7720_get_control,
	.set_control		= ov7720_set_control,
	.get_chip_id		= ov7720_get_chip_id,
#ifdef CONFIG_VIDEO_ADV_DEBUG
	.get_register		= ov7720_get_register,
	.set_register		= ov7720_set_register,
#endif
};

static int ov7720_get_control(struct soc_camera_device *icd,
			       struct v4l2_control *ctrl)
{
	int ret;
	int data, sign;

	ret = 0;

	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		data = reg_read(icd, OV7720_BRIGHT);
		if (data < 0)
			return -EIO;
		sign = reg_read(icd, OV7720_SIGN);
		if (sign < 0)
			return -EIO;
		if (sign & OV7720_SIGN_BRIGHT)
		  ctrl->value = 256 - data;
		else
		  ctrl->value = 256 + data;
		break;
	case V4L2_CID_SATURATION:
		data = reg_read(icd, OV7720_USAT);
		if (data < 0)
			return -EIO;
		ctrl->value = data;
		break;
	case V4L2_CID_AUTO_WHITE_BALANCE:
		data = reg_read(icd, OV7720_COM8);
		if (data < 0)
			return -EIO;
		ctrl->value =  !!(data & OV7720_COM8_AWB_EN);
		break;
/* 	case V4L2_CID_DO_WHITE_BALANCE: */
/* 		/\* not implemented: is a button *\/ */
/* 		break; */
	case V4L2_CID_RED_BALANCE:
		data = reg_read(icd, OV7720_RED);
		if (data < 0)
			return -EIO;
		ctrl->value = data;
		break;
	case V4L2_CID_BLUE_BALANCE:
		data = reg_read(icd, OV7720_BLUE);
		if (data < 0)
			return -EIO;
		ctrl->value = data;
		break;
	case V4L2_CID_GAMMA:
		data = reg_read(icd, OV7720_DSP_CTRL1);
		if (data < 0)
			return -EIO;
		ctrl->value = !!(data & OV7720_DSP1_GAMMA_ON);
		break;
	case V4L2_CID_EXPOSURE:
		data = reg_read(icd, OV7720_AECH);
		if (data < 0)
			return -EIO;
		ctrl->value = data<<8;
		data = reg_read(icd, OV7720_AEC);
		if (data < 0)
			return -EIO;
		ctrl->value += data;
		break;
	case V4L2_CID_AUTOGAIN:
		data = reg_read(icd, OV7720_COM8);
		if (data < 0)
			return -EIO;
		ctrl->value = !!(data & OV7720_COM8_AGC_EN);
		break;
	case V4L2_CID_GAIN:
		data = reg_read(icd, OV7720_AGC);
		if (data < 0)
			return -EIO;
		ctrl->value = data;
		break;
	case V4L2_CID_SHARPNESS:
		data = reg_read(icd, OV7720_EDGE0);
		if (data < 0)
			return -EIO;
		/* this controls magnitude only */
		ctrl->value = data & OV7720_EDGE0_STRENGTH;
		break;
	case V4L2_CID_EXPOSURE_AUTO:
		data = reg_read(icd, OV7720_COM8);
		if (data < 0)
			return -EIO;
		ctrl->value = !!(data & OV7720_COM8_AEC_EN);
		break;
	case OV7720_CID_QVGA:
		data = reg_read(icd, OV7720_COM7);
		if (data < 0)
			return -EIO;
		ctrl->value = !!(data & OV7720_COM7_QVGA);
		break;
	default :
		ret = -EINVAL;
	}

	return ret;
}

static int ov7720_set_control(struct soc_camera_device *icd,
			       struct v4l2_control *ctrl)
{
	int ret;
	int data;
	int shift;
	const struct v4l2_queryctrl *qctrl;

	qctrl = soc_camera_find_qctrl(&ov7720_ops, ctrl->id);

	if (!qctrl)
		return -EINVAL;

	if (ctrl->value > qctrl->maximum || ctrl->value < qctrl->minimum)
		return -EINVAL;

	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		if (ctrl->value > 256) {
			ret = reg_write(icd, OV7720_BRIGHT,
					(u8) ctrl->value - 256);
			if (ret < 0)
				return -EIO;
			ret = reg_clear(icd, OV7720_SIGN, OV7720_SIGN_BRIGHT);
			if (ret < 0)
				return -EIO;
		}
		else {
			ret = reg_write(icd, OV7720_BRIGHT,
					256 - (u8) ctrl->value);
			if (ret < 0)
				return -EIO;
			ret = reg_set(icd, OV7720_SIGN, OV7720_SIGN_BRIGHT);
			if (ret < 0)
				return -EIO;
		}
		break;
	case V4L2_CID_SATURATION:
		ret = reg_write(icd, OV7720_USAT, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		ret = reg_write(icd, OV7720_VSAT, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_AUTO_WHITE_BALANCE:
		if (ctrl->value)
			ret = reg_set(icd, OV7720_COM8, OV7720_COM8_AWB_EN);
		else
			ret = reg_clear(icd, OV7720_COM8, OV7720_COM8_AWB_EN);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_DO_WHITE_BALANCE:
 		/* turn AWB on then off after a while */
		ret = reg_set(icd, OV7720_COM8, OV7720_COM8_AWB_EN);
		if (ret < 0)
			return -EIO;
		msleep(500);
		ret = reg_clear(icd, OV7720_COM8, OV7720_COM8_AWB_EN);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_RED_BALANCE:
		ret = reg_write(icd, OV7720_RED, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_BLUE_BALANCE:
		ret = reg_write(icd, OV7720_BLUE, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_GAMMA:
		if (ctrl->value)
			ret = reg_set(icd,
				      OV7720_DSP_CTRL1, OV7720_DSP1_GAMMA_ON);
		else
			ret = reg_clear(icd,
					OV7720_DSP_CTRL1, OV7720_DSP1_GAMMA_ON);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_EXPOSURE:
		ret = reg_write(icd, OV7720_AEC, (u8) (ctrl->value & 0xff));
		if (ret < 0)
			return -EIO;
		ret = reg_write(icd, OV7720_AECH,
				(u8) ((ctrl->value>>8) & 0xff));
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_AUTOGAIN:
		if (ctrl->value)
			ret = reg_set(icd, OV7720_COM8, OV7720_COM8_AGC_EN);
		else
			ret = reg_clear(icd, OV7720_COM8, OV7720_COM8_AGC_EN);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_GAIN:
		ret = reg_write(icd, OV7720_AGC, (u8) ctrl->value);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_SHARPNESS:
		data = reg_read(icd, OV7720_EDGE0);
		if (data < 0)
			return -EIO;
		/* this controls magnitude only */
		data = (data & ~OV7720_EDGE0_STRENGTH) | ctrl->value;
		ret = reg_write(icd, OV7720_EDGE0, (u8) data);
		if (ret < 0)
			return -EIO;
		break;
	case V4L2_CID_EXPOSURE_AUTO:
		if (ctrl->value)
			ret = reg_set(icd, OV7720_COM8, OV7720_COM8_AEC_EN);
		else
			ret = reg_clear(icd, OV7720_COM8, OV7720_COM8_AEC_EN);
		if (ret < 0)
			return -EIO;
		break;
	case OV7720_CID_QVGA:
		if (ctrl->value) {
			/* switch to QVGA output format */
			ret = reg_set(icd, OV7720_COM7, OV7720_COM7_QVGA);
			shift = 1;
		}
		else {
			/* switch to VGA output format */
			ret = reg_clear(icd, OV7720_COM7, OV7720_COM7_QVGA);
			shift = 0;
		}
		/* update capabilities */
		icd->width_def  = OV7720_W_DEF >> shift;
		icd->height_def = OV7720_H_DEF >> shift;
		icd->x_def      = 0;
		icd->y_def      = 0;
		icd->width      = icd->width_def;
		icd->height     = icd->height_def;
		icd->x_current	= icd->x_def;
		icd->y_current	= icd->y_def;
		icd->width_min  = OV7720_W_MIN >> shift;
		icd->width_max  = OV7720_W_MAX >> shift;
		icd->height_min	= OV7720_H_MIN >> shift;
		icd->height_max	= OV7720_H_MAX >> shift;
		icd->x_min	= 0;
		icd->y_min	= 0;

		/* When changing this control, application must also
		   create or select correct buffers before image acquisition */
		if (ret < 0)
			return -EIO;
		break;
	default :
		ret = -EINVAL;
	}
	return 0;
}

static int ov7720_video_probe(struct soc_camera_device *icd)
{
	int ret;
	dev_dbg(&icd->dev, "ov7720_video_probe begin\n");

	if (!icd->dev.parent ||
	    to_soc_camera_host(icd->dev.parent)->nr != icd->iface)
		return -ENODEV;

	/* everything is OK, sensor was probed in ov7720_detect_client */

	ret = soc_camera_video_start(icd);
	if (ret < 0)
		return ret;

	dev_info(&icd->dev, "Detected an OV7720 chip\n");

	dev_dbg(&icd->dev, "ov7720_video_probe end\n");
	return 0;
}

static void ov7720_video_remove(struct soc_camera_device *icd)
{
	struct ov7720 *ov7720 = container_of(icd, struct ov7720, icd);

	/* turn sensor off */
	ov7720_power(icd, 0);

	dev_dbg(&icd->dev, "Video %x removed: %p, %p\n", ov7720->client->addr,
		icd->dev.parent, icd->vdev);
	soc_camera_video_stop(icd);
}


/*
 *****************************************************************************
 *
 *	platform_data Driver Interface    (get I2C address from platform data)
 *
 *****************************************************************************
 */
static int __init ov7720_platform_probe(struct platform_device *pdev)
{
	struct ov7720_platform_data *pdata;

	pdata = pdev->dev.platform_data;
	if (!pdev || !pdata) {
	  return -ENODEV;
	}

	normal_addr[0] = pdata->i2c_addr;
	ov7720_ctx.icl = &pdata->iclink;

	return 0;
}

static int ov7720_platform_remove(struct platform_device *pdev)
{
	return 0;
}

static struct platform_driver ov7720_platform_driver = {
	.probe  = ov7720_platform_probe,
	.remove = ov7720_platform_remove,
	.driver = {
		.owner = THIS_MODULE,
		.name  = OV7720_DRV_NAME,
	},
};

/*
 *****************************************************************************
 *
 *	I2C Driver Interface
 *
 *****************************************************************************
 */
static int ov7720_detect_client (struct i2c_adapter *adap, int addr, int kind)
{
	s32 data_h, data_l;
	struct i2c_client *client;
	struct soc_camera_device *icd;
	int rc;

	printk(KERN_DEBUG "ov7720_detect_client begin\n");

	if (!i2c_check_functionality(adap, I2C_FUNC_I2C
			| I2C_FUNC_SMBUS_BYTE_DATA))
		return -EIO;

	client = kzalloc(sizeof(struct i2c_client), GFP_KERNEL);
	if (!client)
		return -ENOMEM;

	strlcpy(client->name, "OV7720", I2C_NAME_SIZE);
	client->addr = addr;
	client->adapter = adap;
	client->driver = &ov7720_i2c_driver;

	if ((rc = i2c_attach_client(client)))
		goto attach_err;

	ov7720_ctx.client = client;
	i2c_set_clientdata(client, &ov7720_ctx);

	icd = &ov7720_ctx.icd;

	/* Read out the chip manufacture ID registers */
	data_h = reg_read(icd, OV7720_MIDH); /* dummy initial read */
	data_h = reg_read(icd, OV7720_MIDH);
	data_l = reg_read(icd, OV7720_MIDL);
	if (data_h != OV7720_MIDH_VAL || data_l != OV7720_MIDL_VAL) {
		rc = -ENODEV;
		dev_info(&icd->dev, "No OV7720 detected, MIDH=0x%x, MIDL=0x%x\n",
			 data_h, data_l);
		goto detach_client;
	}
	
	/* turn sensor off until activation */
	rc = ov7720_power(icd, 0);
	if (rc < 0)
		goto detach_client;

	icd->probe	= ov7720_video_probe;
	icd->remove	= ov7720_video_remove;
	icd->ops	= &ov7720_ops;
	icd->control	= &client->dev;
	icd->width_def  = OV7720_W_DEF;
	icd->height_def = OV7720_H_DEF;
	icd->x_def      = 0;
	icd->y_def      = 0;
	icd->width      = icd->width_def;
	icd->height     = icd->height_def;
	icd->x_current	= icd->x_def;
	icd->y_current	= icd->y_def;
	icd->x_min	= 0;
	icd->y_min	= 0;
	icd->width_min	= OV7720_W_MIN;
	icd->width_max	= OV7720_W_MAX;
	icd->height_min	= OV7720_H_MIN;
	icd->height_max	= OV7720_H_MAX;
	icd->y_skip_top	= 0;
	icd->iface	= ov7720_ctx.icl->bus_id;
	ov7720_ctx.datawidth = 8;
	ov7720_ctx.model = V4L2_IDENT_OV7720;

	rc = soc_camera_device_register(icd);
	if (rc)
		goto detach_client;

	printk(KERN_DEBUG "ov7720_detect_client end\n");

	return 0;

detach_client:
	i2c_detach_client(client);
attach_err:
	kfree(client);
	return rc;
}

static int ov7720_attach(struct i2c_adapter *adap)
{
	printk("ov7720_attach begin\n");
	return i2c_probe(adap, &addr_data, ov7720_detect_client);
	printk("ov7720_attach end\n");
}

static int ov7720_detach(struct i2c_client *client)
{
	struct ov7720 *ov7720 = i2c_get_clientdata(client);
	int rc;

	dev_info(&client->dev, "ov7720_detach\n");

	soc_camera_device_unregister(&ov7720->icd);

	rc = i2c_detach_client(client);
	if (rc == 0) {
		kfree(client);
	}

	return rc;
}

static struct i2c_driver ov7720_i2c_driver = {
	.driver = {
		.name = "ov7720",
	},
	.id 		= I2C_DRIVERID_OV7720,
/* 	.class 		= I2C_CLASS_CAM_DIGITAL, */
	.attach_adapter = ov7720_attach,
	.detach_client	= ov7720_detach,
/* 	.probe		= ov7720_probe, */
/* 	.remove		= ov7720_remove, */
};

static int __init ov7720_mod_init(void)
{
	int ret;
	printk("ov7720_mod_init begin\n");
	ret = platform_driver_register(&ov7720_platform_driver);
	if (ret == 0)
		ret = i2c_add_driver(&ov7720_i2c_driver);
	printk("ov7720_mod_init end\n");

	return ret;
}

static void __exit ov7720_mod_exit(void)
{
	i2c_del_driver(&ov7720_i2c_driver);
	platform_driver_unregister(&ov7720_platform_driver);
}

module_init(ov7720_mod_init);
module_exit(ov7720_mod_exit);

MODULE_DESCRIPTION("Omnivision OV7720 Camera driver with soc_camera bus");
MODULE_AUTHOR("Florian Pantaleao <fpantaleao@mobisensesystems.com>");
MODULE_LICENSE("GPL");
