/*
 *  linux/drivers/misc/ucb1x00-adc.c
 *
 *  Copyright (C) 2008 Florian Pantaleao, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This small ADC driver gives user space access to ADC channels.
 * It is inspired by Russel King ucb1x00-ts.c touchscreen driver.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/sched.h>
#include <linux/completion.h>
#include <linux/delay.h>
#include <linux/string.h>
#include <linux/pm.h>
#include <linux/poll.h>
#include <linux/device.h>

#include "ucb1x00.h"

#include <linux/fs.h>
#include <linux/miscdevice.h>

struct ucb1x00_adc {
	struct ucb1x00		*ucb;
	int			use_count;
};

static struct ucb1x00_adc ucbadc;

static int ucb_adc_channel_lut[MISC_UCB_ADC_CHANNEL_MAX+1] = {
  UCB_ADC_INP_AD0, UCB_ADC_INP_AD1, UCB_ADC_INP_AD2, UCB_ADC_INP_AD3
};

/*
 * User space driver interface.
 */

static int ucb1x00_adc_open(struct inode *inode, struct file *filp)
{
  printk("ucb1x00_adc_open\n");
	struct ucb1x00_adc *adc = &ucbadc;
	int ret = 0;

	adc->use_count++;
	filp->private_data = adc;

	return ret;
}

static int ucb1x00_adc_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	struct ucb1x00_adc *adc = &ucbadc;
	int ret = 0;
	
	switch(cmd)
	{
	case MISCIOCADC: {
	        struct misc_adc adc_data;
                int channel;
		unsigned int ii;

		if (copy_from_user(&adc_data, (void*) arg, sizeof(struct misc_adc))) {
		  ret = -EFAULT;
		  break;
                }

		if (adc_data.channel_count > MISC_UCB_NB_ADC_CHANNEL) {
		  ret = -EINVAL;
		  break;
		}

		ucb1x00_adc_enable(adc->ucb);
		for (ii=0; ii<adc_data.channel_count; ii++) {
		  channel = adc_data.channel[ii].number;
		  if ((channel < MISC_UCB_ADC_CHANNEL_MIN) ||
		      (channel > MISC_UCB_ADC_CHANNEL_MAX)) {
		    ret = -EINVAL;
		    break;
		  }
		  adc_data.channel[ii].value = ucb1x00_adc_read(
		      adc->ucb, ucb_adc_channel_lut[channel], UCB_NOSYNC);
		}
		ucb1x00_adc_disable(adc->ucb);

		if (copy_to_user((void*) arg, &adc_data, sizeof(struct misc_adc)))
		  ret = -EFAULT;

		break;
	}

	default:
		ret = -EINVAL;
	}

	return ret;
}

/*
 * Release ADC resources.  Nothing worth to be done here.
 */
static int ucb1x00_adc_release(struct inode *inode, struct file *filp)
{
	struct ucb1x00_adc *adc = filp->private_data;

	adc->use_count--;

	return 0;
}

static struct file_operations ucb1x00_adc_fops = {
	owner:		THIS_MODULE,
	open:		ucb1x00_adc_open,
	release:	ucb1x00_adc_release,
	ioctl:		ucb1x00_adc_ioctl,
};

/*
 * This is an unofficial UCB1x00 ADC considered as a miscdevice:
 *   10 char                Non-serial mice, misc features
 *   16 = /dev/ucb1x00-adc  UCB 1x00 ADC
 */
static struct miscdevice ucb1x00_adc_dev = {
	minor:	UCB1X00_ADC_MINOR,
	name:	"ucb1x00-adc",
	fops:	&ucb1x00_adc_fops,
};


/*
 * Initialization.
 */
static int ucb1x00_adc_add(struct ucb1x00_dev *dev)
{
	ucbadc.ucb = dev->ucb;
	return misc_register(&ucb1x00_adc_dev);
}

static void ucb1x00_adc_remove(struct ucb1x00_dev *dev)
{
	misc_deregister(&ucb1x00_adc_dev);
}

static struct ucb1x00_driver ucb1x00_adc_driver = {
	.add		= ucb1x00_adc_add,
	.remove		= ucb1x00_adc_remove,
};

static int __init ucb1x00_adc_init(void)
{
	return ucb1x00_register_driver(&ucb1x00_adc_driver);
}

static void __exit ucb1x00_adc_exit(void)
{
	ucb1x00_unregister_driver(&ucb1x00_adc_driver);
}

module_init(ucb1x00_adc_init);
module_exit(ucb1x00_adc_exit);

MODULE_AUTHOR("Florian Pantaleao <fpantaleao@mobisensesystems.com>");
MODULE_DESCRIPTION("UCB1x00 user ADC driver");
MODULE_LICENSE("GPL");
