/*
 * Map driver for the Colibri
 * $Id: colibri.c,v 1.3 2006/06/25 14:48:22 cvs Exp $
 *
 * Modified from drivers/mtd/maps/iq80310.c
 * Orignal Author:	Nicolas Pitre
 * Colibri modifications: Detlef Vollmann, Toradex
 *
 * Copyright:   (C) 2001 MontaVista Software Inc.
 * Copyright (c) 2005 Detlef Vollmann, Toradex AG, <dv@vollmann.ch>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>


#define ROM_ADDR 0x00000000
#define WINDOW_SIZE 	32*1024*1024

#ifdef CONFIG_PM
struct pm_dev       *flash_pmdev;
#endif

static struct map_info colibri_map = {
	.name = 		"Colibri flash",
	.size = 		WINDOW_SIZE,
    .bankwidth =    4,
	.phys = 		ROM_ADDR
};

static struct mtd_partition colibri_partitions[] = {
	{
		name:		"Bootloader",
		size:		0x00040000, /* 256K U-Boot */
		offset:		0,
		/*mask_flags:	MTD_WRITEABLE   force read-only */
	},{
		name:		"Boot Configuration",
		size:		0x00040000, /* 256K config params */
		offset:		0x00040000,
		/*mask_flags:	MTD_WRITEABLE   force read-only */
	},{
		name:		"Kernel",
		size:		0x00400000, /* 4M for kernel */
		offset:		0x00080000,
	},{
		name:		"Filesystem",
		size:		MTDPART_SIZ_FULL,
		offset:		0x00480000
	}
};

static struct mtd_info *mymtd;

#ifdef COLIBRI_PARTS_PARSING
static struct mtd_partition *parsed_parts;
static int nr_parsed_parts;

static const char* probes[] = { "RedBoot", "cmdlinepart", NULL };
#endif

#include <linux/mtd/cfi.h>

static int __init init_colibri(void)
{
#ifdef CONFIG_MTD_COLIBRI_P30
	int i, j;
#endif
	printk( "Probing Colibri flash at physical address 0x%08lx (%d-bit buswidth)\n",
		colibri_map.phys, colibri_map.bankwidth * 8 );

	colibri_map.virt = ioremap(colibri_map.phys, colibri_map.size);
	if (!colibri_map.virt) {
		printk("Failed to ioremap\n");
		return -EIO;
	}

	simple_map_init(&colibri_map);

	mymtd = do_map_probe("cfi_probe", &colibri_map);
	if (!mymtd) {
		iounmap((void *)colibri_map.virt);
		return -ENXIO;
	}
	mymtd->owner = THIS_MODULE;

#ifdef CONFIG_MTD_COLIBRI_P30 
	/* Unlock the flash device. */
	printk("Unlocking p30 flash... ");
	for (i = 0; i < mymtd->numeraseregions; i++) {
		for( j = 0; j < mymtd->eraseregions[i].numblocks; j++) {
			mymtd->unlock(mymtd, mymtd->eraseregions[i].offset +
				      j * mymtd->eraseregions[i].erasesize,
				      mymtd->eraseregions[i].erasesize);
		}
	}
	printk("done.\n");
#endif

#ifdef COLIBRI_PARTS_PARSING
	ret = parse_mtd_partitions(mymtd, probes, &parsed_parts, 0);

	if (ret > 0)
		nr_parsed_parts = ret;
#endif

	if (!mymtd) {
		printk(KERN_WARNING "%s is absent.  Skipping\n", colibri_map.name);
#ifdef COLIBRI_PARTS_PARSING
	} else if (nr_parsed_parts) {
		add_mtd_partitions(mymtd, parsed_parts, nr_parsed_parts);
#endif
	} else {
		add_mtd_partitions(mymtd, colibri_partitions, ARRAY_SIZE(colibri_partitions));
	}

	return 0;
}

static void __exit cleanup_colibri(void)
{
	if (mymtd) {
#ifdef COLIBRI_PARTS_PARSING
		if (nr_parsed_parts)
			del_mtd_partitions(mymtd);
		else
#endif
			del_mtd_device(mymtd);

		map_destroy(mymtd);
#ifdef COLIBRI_PARTS_PARSING
		if (parsed_parts)
			kfree(parsed_parts);
#endif
	}

	if (colibri_map.virt)
		iounmap((void *)colibri_map.virt);
}

module_init(init_colibri);
module_exit(cleanup_colibri);

